From d4faafda6f6373bf3e0f01e070535a3ce392f54e Mon Sep 17 00:00:00 2001
From: Shveta Malik <shveta.malik@gmail.com>
Date: Fri, 8 Sep 2023 10:04:04 +0530
Subject: [PATCH v16_2 2/2] Add logical slot sync capability to physical
 standby

This patch attempts to implement logical replication slots synchronization
from primary server to physical standby so that logical subscribers are not
blocked after failover. Now-on, all the logical replication slots created on
primary (assuming configurations are appropriate) are automatically created
on physical standbys and are synced periodically. This has been acheived
by starting slot-sync worker(s) on standby server which pings primary at
regular intervals to get the logical slots information and create/update the
slots locally.

For max number of slot-sync workers on standby, new GUC max_slotsync_workers
has been added, default value and max value is kept at 2 and 50 respectively.
This parameter can only be set at server start.

Now replication launcher on physical standby queries primary to get list
of dbids which belong to slots mentioned in GUC 'synchronize_slot_names'.
Once it gets the dbids, if dbids < max_slotsync_workers, it starts only
that many workers and if dbids > max_slotsync_workers, it starts
max_slotsync_workers and divides the work equally among them. Each worker
is then responsible to keep on syncing the concerned logical slots belonging
to the DBs assigned to it.

Let us say slots mentioned in 'synchronize_slot_names' on primary belongs to
4 DBs and say 'max_slotsync_workers' is 4, then a new worker will be launched
for each db. If a new logical slot with a different DB is found by replication
launcher, it will assign this new db to the worker handling the minimum number
of dbs currently (or first worker in case of equal count).

Each slot-sync worker will have its own dbids list. Since the upper limit
of this dbid-count is not known, it needs to be handled using dsa. We
initially allocate memory to hold 100 dbids for each worker. If this limit
is exhausted, we reallocate this memory with size incremented again by 100.

The naptime of worker is tuned as per the activity on primary. Each worker
starts with naptime of 10ms and if no activity is observed on primary for some
time, then naptime is increased to 10sec. And if activity is observed again,
naptime is reduced back to 10ms. Each worker does it by choosing one slot
(first one assigned to it) for monitoring purpose. If there is no change
in lsn of that slot for say over 10 sync-checks, naptime is increased to 10sec
and as soon as a change is observed, naptime is reduced back to 10ms.

The logical slots created by slot-sync workers on physical standbys are
not allowed to be consumed. Any attempt to do pg_logical_slot_get_changes
on such slots will result in an error.

If a logical slot is invalidated on primary, slot on standby is also
invalidated. If a logical slot on primary is valid but is invalidated
on standby due to conflict (say required rows removed on primary), then
that slot is dropped and recreated on standby in next sync-cycle. It is okay
to recerate such slots as long as these are not consumable on standby (which
is the case currently).

If there is any change in synchronize_slot_names, then the slots which are no
longer part of it or the ones which no longer exist on primary will be dropped
by slot-sync workers on physical standbys.
---
 doc/src/sgml/config.sgml                      |  21 +
 src/backend/postmaster/bgworker.c             |   3 +
 .../libpqwalreceiver/libpqwalreceiver.c       |  77 ++
 src/backend/replication/logical/Makefile      |   1 +
 src/backend/replication/logical/launcher.c    | 833 +++++++++++++--
 .../replication/logical/logicalfuncs.c        |  13 +
 src/backend/replication/logical/meson.build   |   1 +
 src/backend/replication/logical/slotsync.c    | 972 ++++++++++++++++++
 src/backend/replication/logical/tablesync.c   |   1 +
 src/backend/replication/repl_gram.y           |  32 +-
 src/backend/replication/repl_scanner.l        |   2 +
 src/backend/replication/slot.c                |   3 +-
 src/backend/replication/slotfuncs.c           |  28 +
 src/backend/replication/walsender.c           | 104 ++
 src/backend/storage/lmgr/lwlock.c             |   2 +
 src/backend/storage/lmgr/lwlocknames.txt      |   1 +
 .../utils/activity/wait_event_names.txt       |   1 +
 src/backend/utils/misc/guc_tables.c           |  16 +
 src/backend/utils/misc/postgresql.conf.sample |   1 +
 src/include/catalog/pg_proc.dat               |   4 +
 src/include/commands/subscriptioncmds.h       |   4 +
 src/include/nodes/replnodes.h                 |   9 +
 src/include/postmaster/bgworker_internals.h   |   1 +
 src/include/replication/logicallauncher.h     |   4 +
 src/include/replication/logicalworker.h       |   1 +
 src/include/replication/slot.h                |   7 +-
 src/include/replication/walreceiver.h         |  23 +
 src/include/replication/worker_internal.h     |  49 +-
 src/include/storage/lwlock.h                  |   1 +
 src/tools/pgindent/typedefs.list              |   4 +
 30 files changed, 2148 insertions(+), 71 deletions(-)
 create mode 100644 src/backend/replication/logical/slotsync.c
 mode change 100644 => 100755 src/backend/replication/walsender.c

diff --git a/doc/src/sgml/config.sgml b/doc/src/sgml/config.sgml
index 3d81bd308c..b6c1948fdf 100644
--- a/doc/src/sgml/config.sgml
+++ b/doc/src/sgml/config.sgml
@@ -5015,6 +5015,27 @@ ANY <replaceable class="parameter">num_sync</replaceable> ( <replaceable class="
       </listitem>
      </varlistentry>
 
+     <varlistentry id="guc-max-slotsync-workers" xreflabel="max_slotsync_workers">
+      <term><varname>max_slotsync_workers</varname> (<type>integer</type>)
+      <indexterm>
+       <primary><varname>max_slotsync_workers</varname> configuration parameter</primary>
+      </indexterm>
+      </term>
+      <listitem>
+       <para>
+        Specifies maximum number of slot synchronization workers.
+       </para>
+       <para>
+        Slot synchronization workers are taken from the pool defined by
+        <varname>max_worker_processes</varname>.
+       </para>
+       <para>
+        The default value is 2. This parameter can only be set at server
+        start.
+       </para>
+      </listitem>
+     </varlistentry>
+
      </variablelist>
     </sect2>
 
diff --git a/src/backend/postmaster/bgworker.c b/src/backend/postmaster/bgworker.c
index 505e38376c..216287d56a 100644
--- a/src/backend/postmaster/bgworker.c
+++ b/src/backend/postmaster/bgworker.c
@@ -129,6 +129,9 @@ static const struct
 	{
 		"ApplyWorkerMain", ApplyWorkerMain
 	},
+	{
+		"ReplSlotSyncMain", ReplSlotSyncMain
+	},
 	{
 		"ParallelApplyWorkerMain", ParallelApplyWorkerMain
 	},
diff --git a/src/backend/replication/libpqwalreceiver/libpqwalreceiver.c b/src/backend/replication/libpqwalreceiver/libpqwalreceiver.c
index 60d5c1fc40..ea60b2969d 100644
--- a/src/backend/replication/libpqwalreceiver/libpqwalreceiver.c
+++ b/src/backend/replication/libpqwalreceiver/libpqwalreceiver.c
@@ -34,6 +34,7 @@
 #include "utils/memutils.h"
 #include "utils/pg_lsn.h"
 #include "utils/tuplestore.h"
+#include "utils/varlena.h"
 
 PG_MODULE_MAGIC;
 
@@ -58,6 +59,8 @@ static void libpqrcv_get_senderinfo(WalReceiverConn *conn,
 									char **sender_host, int *sender_port);
 static char *libpqrcv_identify_system(WalReceiverConn *conn,
 									  TimeLineID *primary_tli);
+static List *libpqrcv_get_dbinfo_for_logical_slots(WalReceiverConn *conn,
+												const char *slot_names);
 static int	libpqrcv_server_version(WalReceiverConn *conn);
 static void libpqrcv_readtimelinehistoryfile(WalReceiverConn *conn,
 											 TimeLineID tli, char **filename,
@@ -96,6 +99,7 @@ static WalReceiverFunctionsType PQWalReceiverFunctions = {
 	.walrcv_receive = libpqrcv_receive,
 	.walrcv_send = libpqrcv_send,
 	.walrcv_create_slot = libpqrcv_create_slot,
+	.walrcv_get_dbinfo_for_logical_slots = libpqrcv_get_dbinfo_for_logical_slots,
 	.walrcv_get_backend_pid = libpqrcv_get_backend_pid,
 	.walrcv_exec = libpqrcv_exec,
 	.walrcv_disconnect = libpqrcv_disconnect
@@ -409,6 +413,79 @@ libpqrcv_server_version(WalReceiverConn *conn)
 	return PQserverVersion(conn->streamConn);
 }
 
+/*
+ * Get DB info for logical slots
+ *
+ * It gets the DBIDs for slot_names from primary. The list obatined has no
+ * duplicacy of DBIds.
+ */
+static List *
+libpqrcv_get_dbinfo_for_logical_slots(WalReceiverConn *conn,
+									const char *slot_names)
+{
+	PGresult   *res;
+	List	   *slotlist = NIL;
+	int			ntuples;
+	StringInfoData s;
+	WalRcvRepSlotDbData *slot_data;
+
+	initStringInfo(&s);
+	appendStringInfoString(&s, "LIST_DBID_FOR_LOGICAL_SLOTS");
+
+	if (strcmp(slot_names, "") != 0 && strcmp(slot_names, "*") != 0)
+	{
+		char	   *rawnames;
+		List	   *namelist;
+		ListCell   *lc;
+
+		appendStringInfoChar(&s, ' ');
+		rawnames = pstrdup(slot_names);
+		SplitIdentifierString(rawnames, ',', &namelist);
+		foreach (lc, namelist)
+		{
+			if (lc != list_head(namelist))
+				appendStringInfoChar(&s, ',');
+			appendStringInfo(&s, "%s",
+							 quote_identifier(lfirst(lc)));
+		}
+	}
+
+	res = libpqrcv_PQexec(conn->streamConn, s.data);
+	pfree(s.data);
+	if (PQresultStatus(res) != PGRES_TUPLES_OK)
+	{
+		PQclear(res);
+		ereport(ERROR,
+				(errmsg("could not receive list of slots from the primary server: %s",
+						pchomp(PQerrorMessage(conn->streamConn)))));
+	}
+	if (PQnfields(res) != 1)
+	{
+		int			nfields = PQnfields(res);
+
+		PQclear(res);
+		ereport(ERROR,
+				(errmsg("invalid response from primary server"),
+				 errdetail("Could not get list of slots: got %d fields, "
+							"expected %d fields.",
+						   nfields, 1)));
+	}
+
+	ntuples = PQntuples(res);
+	for (int i = 0; i < ntuples; i++)
+	{
+		slot_data = palloc0(sizeof(WalRcvRepSlotDbData));
+		if (!PQgetisnull(res, i, 0))
+			slot_data->database = atooid(PQgetvalue(res, i, 0));
+
+		slotlist = lappend(slotlist, slot_data);
+	}
+
+	PQclear(res);
+
+	return slotlist;
+}
+
 /*
  * Start streaming WAL data from given streaming options.
  *
diff --git a/src/backend/replication/logical/Makefile b/src/backend/replication/logical/Makefile
index 2dc25e37bb..ba03eeff1c 100644
--- a/src/backend/replication/logical/Makefile
+++ b/src/backend/replication/logical/Makefile
@@ -25,6 +25,7 @@ OBJS = \
 	proto.o \
 	relation.o \
 	reorderbuffer.o \
+	slotsync.o \
 	snapbuild.o \
 	tablesync.o \
 	worker.o
diff --git a/src/backend/replication/logical/launcher.c b/src/backend/replication/logical/launcher.c
index 7882fc91ce..fbc6f7ec2f 100644
--- a/src/backend/replication/logical/launcher.c
+++ b/src/backend/replication/logical/launcher.c
@@ -22,6 +22,7 @@
 #include "access/htup_details.h"
 #include "access/tableam.h"
 #include "access/xact.h"
+#include "catalog/pg_authid.h"
 #include "catalog/pg_subscription.h"
 #include "catalog/pg_subscription_rel.h"
 #include "funcapi.h"
@@ -57,6 +58,21 @@
 int			max_logical_replication_workers = 4;
 int			max_sync_workers_per_subscription = 2;
 int			max_parallel_apply_workers_per_subscription = 2;
+int			max_slotsync_workers = 2;
+
+/*
+ * Initial allocation size for dbids array for each SlotSyncWorker in dynamic
+ * shared memory.
+ */
+#define DB_PER_WORKER_ALLOC_INIT 100
+
+/*
+ * Once initially allocated size is exhausted for dbids array, it is extended by
+ * DB_PER_WORKER_ALLOC_EXTRA size.
+ */
+#define DB_PER_WORKER_ALLOC_EXTRA 100
+
+SlotSyncWorker *MySlotSyncWorker = NULL;
 
 LogicalRepWorker *MyLogicalRepWorker = NULL;
 
@@ -70,6 +86,7 @@ typedef struct LogicalRepCtxStruct
 	dshash_table_handle last_start_dsh;
 
 	/* Background workers. */
+	SlotSyncWorker *ss_workers; /* slot-sync workers */
 	LogicalRepWorker workers[FLEXIBLE_ARRAY_MEMBER];
 } LogicalRepCtxStruct;
 
@@ -931,7 +948,16 @@ ApplyLauncherRegister(void)
 	memset(&bgw, 0, sizeof(bgw));
 	bgw.bgw_flags = BGWORKER_SHMEM_ACCESS |
 		BGWORKER_BACKEND_DATABASE_CONNECTION;
-	bgw.bgw_start_time = BgWorkerStart_RecoveryFinished;
+
+	/*
+	 * The launcher now takes care of launching both logical apply workers and
+	 * logical slot-sync workers. Thus to cater to the requirements of both,
+	 * start it as soon as a consistent state is reached. This will help
+	 * slot-sync workers to start timely on a physical standby while on a
+	 * non-standby server, it holds same meaning as that of
+	 * BgWorkerStart_RecoveryFinished.
+	 */
+	bgw.bgw_start_time = BgWorkerStart_ConsistentState;
 	snprintf(bgw.bgw_library_name, MAXPGPATH, "postgres");
 	snprintf(bgw.bgw_function_name, BGW_MAXLEN, "ApplyLauncherMain");
 	snprintf(bgw.bgw_name, BGW_MAXLEN,
@@ -953,6 +979,7 @@ void
 ApplyLauncherShmemInit(void)
 {
 	bool		found;
+	Size		ssw_size;
 
 	LogicalRepCtx = (LogicalRepCtxStruct *)
 		ShmemInitStruct("Logical Replication Launcher Data",
@@ -977,6 +1004,14 @@ ApplyLauncherShmemInit(void)
 			SpinLockInit(&worker->relmutex);
 		}
 	}
+
+	/* Allocate shared-memory for slot-sync workers pool now */
+	ssw_size = mul_size(max_slotsync_workers, sizeof(SlotSyncWorker));
+	LogicalRepCtx->ss_workers = (SlotSyncWorker *)
+		ShmemInitStruct("Replication slot-sync workers", ssw_size, &found);
+
+	if (!found)
+		memset(LogicalRepCtx->ss_workers, 0, ssw_size);
 }
 
 /*
@@ -1114,6 +1149,731 @@ ApplyLauncherWakeup(void)
 		kill(LogicalRepCtx->launcher_pid, SIGUSR1);
 }
 
+/*
+ * Clean up slot-sync worker info.
+ */
+static void
+slotsync_worker_cleanup(SlotSyncWorker *worker)
+{
+	Assert(LWLockHeldByMeInMode(SlotSyncWorkerLock, LW_EXCLUSIVE));
+
+	worker->in_use = false;
+	worker->proc = NULL;
+	worker->dbid = InvalidOid;
+	worker->userid = InvalidOid;
+	worker->slot = -1;
+
+	if (DsaPointerIsValid(worker->dbids_dp))
+	{
+		dsa_free(worker->dbids_dsa, worker->dbids_dp);
+		worker->dbids_dp = InvalidDsaPointer;
+	}
+
+	if (worker->dbids_dsa)
+	{
+		dsa_detach(worker->dbids_dsa);
+		worker->dbids_dsa = NULL;
+	}
+
+	worker->dbcount = 0;
+
+	MemSet(NameStr(worker->monitoring_info.slot_name), 0, NAMEDATALEN);
+	worker->monitoring_info.inactivity_count = 0;
+	worker->monitoring_info.confirmed_lsn = 0;
+}
+
+/*
+ * Attach Slot-sync worker to worker-slot assigned by launcher.
+ */
+void
+slotsync_worker_attach(int slot)
+{
+	/* Block concurrent access. */
+	LWLockAcquire(SlotSyncWorkerLock, LW_EXCLUSIVE);
+
+	Assert(slot >= 0 && slot < max_slotsync_workers);
+	MySlotSyncWorker = &LogicalRepCtx->ss_workers[slot];
+	MySlotSyncWorker->slot = slot;
+
+	if (!MySlotSyncWorker->in_use)
+	{
+		LWLockRelease(SlotSyncWorkerLock);
+		ereport(ERROR,
+				(errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
+				 errmsg("replication slot-sync worker slot %d is "
+						"empty, cannot attach", slot)));
+	}
+
+	if (MySlotSyncWorker->proc)
+	{
+		LWLockRelease(SlotSyncWorkerLock);
+		ereport(ERROR,
+				(errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
+				 errmsg("replication slot-sync worker slot %d is "
+						"already used by another worker, cannot attach", slot)));
+	}
+
+	MySlotSyncWorker->proc = MyProc;
+
+	LWLockRelease(SlotSyncWorkerLock);
+}
+
+/*
+ * Wait for a slot-sync worker to start up and attach to the shmem context.
+ *
+ * This is only needed for cleaning up the shared memory in case the worker
+ * fails to attach.
+ *
+ * Returns whether the attach was successful.
+ */
+static bool
+WaitForSlotSyncWorkerAttach(SlotSyncWorker *worker,
+							uint16 generation,
+							BackgroundWorkerHandle *handle)
+{
+	BgwHandleStatus status;
+	int			rc;
+
+	for (;;)
+	{
+		pid_t		pid;
+
+		CHECK_FOR_INTERRUPTS();
+
+		LWLockAcquire(SlotSyncWorkerLock, LW_SHARED);
+
+		/* Worker either died or has started. Return false if died. */
+		if (!worker->in_use || worker->proc)
+		{
+			LWLockRelease(SlotSyncWorkerLock);
+			return worker->in_use;
+		}
+
+		LWLockRelease(SlotSyncWorkerLock);
+
+		/* Check if worker has died before attaching, and clean up after it. */
+		status = GetBackgroundWorkerPid(handle, &pid);
+
+		if (status == BGWH_STOPPED)
+		{
+			LWLockAcquire(SlotSyncWorkerLock, LW_EXCLUSIVE);
+			/* Ensure that this was indeed the worker we waited for. */
+			if (generation == worker->generation)
+				slotsync_worker_cleanup(worker);
+			LWLockRelease(SlotSyncWorkerLock);
+			return false;
+		}
+
+		/*
+		 * We need timeout because we generally don't get notified via latch
+		 * about the worker attach.  But we don't expect to have to wait long.
+		 */
+		rc = WaitLatch(MyLatch,
+					   WL_LATCH_SET | WL_TIMEOUT | WL_EXIT_ON_PM_DEATH,
+					   10L, WAIT_EVENT_BGWORKER_STARTUP);
+
+		if (rc & WL_LATCH_SET)
+		{
+			ResetLatch(MyLatch);
+			CHECK_FOR_INTERRUPTS();
+		}
+	}
+}
+
+/*
+ * Slot Sync worker find.
+ *
+ * Walks the slot-sync workers pool and searches for one that matches given
+ * dbid. Since one worker can manage multiple dbs, so it walks the db array in
+ * each worker to find the match.
+ */
+static SlotSyncWorker *
+slotsync_worker_find(Oid dbid)
+{
+	int			i;
+	SlotSyncWorker *res = NULL;
+	Oid		   *dbids;
+
+	Assert(LWLockHeldByMe(SlotSyncWorkerLock));
+
+	/* Search for attached worker for a given dbid */
+	for (i = 0; i < max_slotsync_workers; i++)
+	{
+		SlotSyncWorker *w = &LogicalRepCtx->ss_workers[i];
+		int			cnt;
+
+		if (!w->in_use)
+			continue;
+
+		dbids = (Oid *) dsa_get_address(w->dbids_dsa, w->dbids_dp);
+		SpinLockAcquire(&w->mutex);
+		for (cnt = 0; cnt < w->dbcount; cnt++)
+		{
+			Oid			wdbid = dbids[cnt];
+
+			if (wdbid == dbid)
+			{
+				res = w;
+				break;
+			}
+		}
+		SpinLockRelease(&w->mutex);
+
+		/* if worker is found, break the outer loop */
+		if (res)
+			break;
+	}
+
+	return res;
+}
+
+/*
+ * Setup DSA for slot-sync worker.
+ *
+ * DSA is needed for dbids array. Since max number of dbs a worker can manage
+ * is not known, so initially fixed size to hold DB_PER_WORKER_ALLOC_INIT
+ * dbs is allocated. If this size is exhausted, it can be extended using
+ * dsa free and allocate routines.
+ */
+static dsa_handle
+slot_sync_dsa_setup(SlotSyncWorker *worker, int alloc_db_count)
+{
+	dsa_area   *dbids_dsa;
+	dsa_pointer dbids_dp;
+	dsa_handle	dbids_dsa_handle;
+	MemoryContext oldcontext;
+
+	/* Be sure any memory allocated by DSA routines is persistent. */
+	oldcontext = MemoryContextSwitchTo(TopMemoryContext);
+
+	dbids_dsa = dsa_create(LWTRANCHE_SLOTSYNC_DSA);
+	dsa_pin(dbids_dsa);
+	dsa_pin_mapping(dbids_dsa);
+
+	dbids_dp = dsa_allocate0(dbids_dsa, alloc_db_count * sizeof(Oid));
+
+	/* set-up worker */
+	SpinLockInit(&worker->mutex);
+	worker->dbcount = 0;
+	worker->dbids_dsa = dbids_dsa;
+	worker->dbids_dp = dbids_dp;
+
+	/* Get the handle. This is the one which can be passed to worker processes */
+	dbids_dsa_handle = dsa_get_handle(dbids_dsa);
+
+	ereport(DEBUG1,
+			(errmsg("allocated dsa for slot-sync worker for dbcount: %d",
+					alloc_db_count)));
+
+	MemoryContextSwitchTo(oldcontext);
+
+	return dbids_dsa_handle;
+}
+
+/*
+ * Stop the slot-sync worker and wait until it detaches from the
+ * slot.
+ */
+static void
+slot_sync_worker_stop(SlotSyncWorker *worker)
+{
+
+	Assert(LWLockHeldByMeInMode(SlotSyncWorkerLock, LW_SHARED));
+
+	/* send SIGINT so that it exists cleanly ... */
+	kill(worker->proc->pid, SIGINT);
+
+	/* ... and wait for it to exit. */
+	for (;;)
+	{
+		int			rc;
+
+		/* is it gone? */
+		if (!worker->proc)
+			break;
+
+		LWLockRelease(SlotSyncWorkerLock);
+
+		/* Wait a bit --- we don't expect to have to wait long. */
+		rc = WaitLatch(MyLatch,
+					   WL_LATCH_SET | WL_TIMEOUT | WL_EXIT_ON_PM_DEATH,
+					   10L, WAIT_EVENT_BGWORKER_SHUTDOWN);
+
+		if (rc & WL_LATCH_SET)
+		{
+			ResetLatch(MyLatch);
+			CHECK_FOR_INTERRUPTS();
+		}
+
+		LWLockAcquire(SlotSyncWorkerLock, LW_SHARED);
+	}
+
+}
+
+/*
+ * Slot sync worker launch or reuse
+ *
+ * Start new slot-sync background worker from the pool of available workers
+ * going by max_slotsync_workers count. If the worker pool is exhausted,
+ * reuse the existing worker with minimum number of dbs. The idea is to
+ * always distribute the dbs equally among launched workers.
+ * If initially allocated dbids array is exhausted for the selected worker,
+ * reallocate the dbids array with increased size and copy the existing
+ * dbids to it and assign the new one as well.
+ *
+ * Returns true on success, false on failure.
+ */
+static bool
+slot_sync_worker_launch_or_reuse(Oid dbid, Oid userid)
+{
+	BackgroundWorker bgw;
+	BackgroundWorkerHandle *bgw_handle;
+	uint16		generation;
+	SlotSyncWorker *worker = NULL;
+	uint32		mindbcnt = 0;
+	uint32		alloc_count = 0;
+	uint32		copied_dbcnt = 0;
+	Oid		   *copied_dbids = NULL;
+	int			worker_slot = -1;
+	dsa_handle	handle;
+	Oid		   *dbids;
+
+	Assert(OidIsValid(dbid));
+
+	/*
+	 * We need to do the modification of the shared memory under lock so that
+	 * we have consistent view.
+	 */
+	LWLockAcquire(SlotSyncWorkerLock, LW_EXCLUSIVE);
+
+	/* Find unused worker slot. */
+	for (int i = 0; i < max_slotsync_workers; i++)
+	{
+		SlotSyncWorker *w = &LogicalRepCtx->ss_workers[i];
+
+		if (!w->in_use)
+		{
+			worker = w;
+			worker_slot = i;
+			break;
+		}
+	}
+
+	/*
+	 * If all the workers are currently in use. Find the one with minimum
+	 * number of dbs and use that.
+	 */
+	if (!worker)
+	{
+		for (int i = 0; i < max_slotsync_workers; i++)
+		{
+			SlotSyncWorker *w = &LogicalRepCtx->ss_workers[i];
+
+			if (i == 0)
+			{
+				mindbcnt = w->dbcount;
+				worker = w;
+				worker_slot = i;
+			}
+			else if (w->dbcount < mindbcnt)
+			{
+				mindbcnt = w->dbcount;
+				worker = w;
+				worker_slot = i;
+			}
+		}
+	}
+
+	/*
+	 * If worker is being reused, and there is vacancy in dbids array, just
+	 * update dbids array and dbcount and we are done. But if dbids array is
+	 * exhausted, reallocate dbids using dsa and copy the old dbids and assign
+	 * the new one as well.
+	 */
+	if (worker->in_use)
+	{
+		dbids = (Oid *) dsa_get_address(worker->dbids_dsa, worker->dbids_dp);
+
+		SpinLockAcquire(&worker->mutex);
+
+		if (worker->dbcount < DB_PER_WORKER_ALLOC_INIT)
+		{
+			dbids[worker->dbcount++] = dbid;
+
+		}
+		else
+		{
+			MemoryContext oldcontext;
+
+			/* Be sure any memory allocated by DSA routines is persistent. */
+			oldcontext = MemoryContextSwitchTo(TopMemoryContext);
+
+			/* Remember the old dbids before we reallocate dsa. */
+			copied_dbcnt = worker->dbcount;
+			copied_dbids = (Oid *) palloc0(worker->dbcount * sizeof(Oid));
+			memcpy(copied_dbids, dbids, worker->dbcount * sizeof(Oid));
+
+			alloc_count = copied_dbcnt + DB_PER_WORKER_ALLOC_EXTRA;
+
+			/* Free the existing dbids and allocate new with increased size */
+			if (DsaPointerIsValid(worker->dbids_dp))
+				dsa_free(worker->dbids_dsa, worker->dbids_dp);
+
+			worker->dbids_dp = dsa_allocate0(worker->dbids_dsa,
+											 alloc_count * sizeof(Oid));
+
+			dbids = (Oid *) dsa_get_address(worker->dbids_dsa, worker->dbids_dp);
+
+			/* Copy the existing dbids */
+			worker->dbcount = copied_dbcnt;
+			memcpy(dbids, copied_dbids, copied_dbcnt * sizeof(Oid));
+
+			/* Assign new dbid */
+			dbids[worker->dbcount++] = dbid;
+
+			MemoryContextSwitchTo(oldcontext);
+		}
+
+		SpinLockRelease(&worker->mutex);
+		LWLockRelease(SlotSyncWorkerLock);
+
+		ereport(LOG,
+				(errmsg("Adding database %d to replication slot"
+						" synchronization worker %d; dbcount now: %d",
+						dbid, worker_slot, worker->dbcount)));
+
+		return true;
+
+	}
+
+	/* Prepare the new worker. */
+	worker->launch_time = GetCurrentTimestamp();
+	worker->in_use = true;
+
+	/*
+	 * 'proc' and 'slot' will be assigned in ReplSlotSyncMain when we attach
+	 * this worker to a particular worker-pool slot
+	 */
+	worker->proc = NULL;
+	worker->slot = -1;
+
+	/* TODO: do we really need these 2? analyse more here */
+	worker->dbid = dbid;
+	worker->generation++;
+
+	/* Initial DSA setup for dbids array to hold DB_PER_WORKER_ALLOC_INIT dbs */
+	handle = slot_sync_dsa_setup(worker, DB_PER_WORKER_ALLOC_INIT);
+	dbids = (Oid *) dsa_get_address(worker->dbids_dsa, worker->dbids_dp);
+
+	dbids[worker->dbcount++] = dbid;
+	worker->userid = userid;
+
+	/* Before releasing lock, remember generation for future identification. */
+	generation = worker->generation;
+
+	LWLockRelease(SlotSyncWorkerLock);
+
+	/* Register the new dynamic worker. */
+	memset(&bgw, 0, sizeof(bgw));
+	bgw.bgw_flags = BGWORKER_SHMEM_ACCESS |
+		BGWORKER_BACKEND_DATABASE_CONNECTION;
+	bgw.bgw_start_time = BgWorkerStart_ConsistentState;
+	snprintf(bgw.bgw_library_name, MAXPGPATH, "postgres");
+
+	snprintf(bgw.bgw_function_name, BGW_MAXLEN, "ReplSlotSyncMain");
+
+	Assert(worker_slot >= 0);
+	snprintf(bgw.bgw_name, BGW_MAXLEN,
+			 "replication slot-sync worker %d", worker_slot);
+
+	snprintf(bgw.bgw_type, BGW_MAXLEN, "slot-sync worker");
+
+	bgw.bgw_restart_time = BGW_NEVER_RESTART;
+	bgw.bgw_notify_pid = MyProcPid;
+	bgw.bgw_main_arg = Int32GetDatum(worker_slot);
+
+	memcpy(bgw.bgw_extra, &handle, sizeof(dsa_handle));
+
+	if (!RegisterDynamicBackgroundWorker(&bgw, &bgw_handle))
+	{
+		/* Failed to start worker, so clean up the worker slot. */
+		LWLockAcquire(SlotSyncWorkerLock, LW_EXCLUSIVE);
+		Assert(generation == worker->generation);
+		slotsync_worker_cleanup(worker);
+		LWLockRelease(SlotSyncWorkerLock);
+
+		ereport(WARNING,
+				(errcode(ERRCODE_CONFIGURATION_LIMIT_EXCEEDED),
+				 errmsg("out of background worker slots"),
+				 errhint("You might need to increase %s.", "max_worker_processes")));
+		return false;
+	}
+
+	/* Now wait until it attaches. */
+	return WaitForSlotSyncWorkerAttach(worker, generation, bgw_handle);
+}
+
+/*
+ * Slot-sync workers remove obsolete DBs from db-list
+ *
+ * If the DBIds fetched from primary are lesser than the ones being managed by
+ * slot-sync workers, remove extra dbs from worker's db-list. This may happen
+ * if some slots are removed on primary or 'synchronize_slot_names' have been
+ * changed by user.
+ */
+static void
+slot_sync_remove_obsolete_dbs(List *remote_dbs)
+{
+	ListCell   *lc;
+	Oid		   *dbids;
+
+	LWLockAcquire(SlotSyncWorkerLock, LW_EXCLUSIVE);
+
+	/* Traverse slot-sync-workers to validate the DBs */
+	for (int widx = 0; widx < max_slotsync_workers; widx++)
+	{
+		SlotSyncWorker *worker = &LogicalRepCtx->ss_workers[widx];
+
+		if (!worker->in_use)
+			continue;
+
+		dbids = (Oid *) dsa_get_address(worker->dbids_dsa, worker->dbids_dp);
+
+		for (int dbidx = 0; dbidx < worker->dbcount;)
+		{
+			Oid			wdbid = dbids[dbidx];
+			bool		found = false;
+
+			/* Check if current DB is still present in remote-db-list */
+			foreach(lc, remote_dbs)
+			{
+				WalRcvRepSlotDbData *slot_db_data = lfirst(lc);
+
+				if (slot_db_data->database == wdbid)
+				{
+					found = true;
+					break;
+				}
+			}
+
+			/* If not found, then delete this db from worker's db-list */
+			if (!found)
+			{
+				SpinLockAcquire(&worker->mutex);
+
+				for (int i = dbidx; i < worker->dbcount; i++)
+				{
+					/* Shift the DBs and get rid of wdbid */
+					if (i < (worker->dbcount - 1))
+						dbids[i] = dbids[i + 1];
+				}
+
+				worker->dbcount--;
+				SpinLockRelease(&worker->mutex);
+
+				ereport(LOG,
+						(errmsg("Removed database %d from replication slot"
+								" synchronization worker %d",
+								wdbid, worker->slot)));
+			}
+			/* Else move to next db-position */
+			else
+			{
+				dbidx++;
+			}
+		}
+	}
+
+	LWLockRelease(SlotSyncWorkerLock);
+
+	/* If dbcount for any worker has become 0, shut it down */
+	for (int widx = 0; widx < max_slotsync_workers; widx++)
+	{
+		SlotSyncWorker *worker = &LogicalRepCtx->ss_workers[widx];
+
+		if (worker->in_use && !worker->dbcount)
+		{
+			int			slot = worker->slot;
+
+			LWLockAcquire(SlotSyncWorkerLock, LW_SHARED);
+			slot_sync_worker_stop(worker);
+			LWLockRelease(SlotSyncWorkerLock);
+
+			LWLockAcquire(SlotSyncWorkerLock, LW_EXCLUSIVE);
+			slotsync_worker_cleanup(worker);
+			LWLockRelease(SlotSyncWorkerLock);
+			ereport(LOG,
+					(errmsg("Stopped replication slot-sync worker %d",
+							slot)));
+		}
+	}
+
+}
+
+/*
+ * Start slot-sync background workers.
+ *
+ * It connects to primary, get the list of DBIDs for slots configured in
+ * synchronize_slot_names. It then launces the slot-sync workers as per
+ * max_slotsync_workers and then assign the DBs equally to the workers
+ * launched.
+ */
+static void
+ApplyLauncherStartSlotSync(long *wait_time)
+{
+	WalReceiverConn *wrconn;
+	char	   *err;
+	List	   *slots_dbs;
+	ListCell   *lc;
+	MemoryContext tmpctx;
+	MemoryContext oldctx;
+
+	if (max_slotsync_workers == 0)
+		return;
+
+	if (strcmp(synchronize_slot_names, "") == 0)
+		return;
+
+	wrconn = walrcv_connect(PrimaryConnInfo, false, false,
+							"Logical Replication Launcher", &err);
+	if (!wrconn)
+		ereport(ERROR,
+				(errmsg("could not connect to the primary server: %s", err)));
+
+	/* Use temporary context for the slot list and worker info. */
+	tmpctx = AllocSetContextCreate(TopMemoryContext,
+								   "Logical Replication Launcher Slot Sync ctx",
+								   ALLOCSET_DEFAULT_SIZES);
+	oldctx = MemoryContextSwitchTo(tmpctx);
+
+	slots_dbs = walrcv_get_dbinfo_for_logical_slots(wrconn,
+													synchronize_slot_names);
+
+	slot_sync_remove_obsolete_dbs(slots_dbs);
+
+	foreach(lc, slots_dbs)
+	{
+		WalRcvRepSlotDbData *slot_db_data = lfirst(lc);
+		SlotSyncWorker *w;
+		TimestampTz last_sync;
+		TimestampTz now;
+		long		elapsed;
+
+		if (!OidIsValid(slot_db_data->database))
+			continue;
+
+		LWLockAcquire(SlotSyncWorkerLock, LW_SHARED);
+		w = slotsync_worker_find(slot_db_data->database);
+		LWLockRelease(SlotSyncWorkerLock);
+
+		if (w != NULL)
+			continue;			/* worker is running already */
+
+		/*
+		 * If the worker is eligible to start now, launch it. Otherwise,
+		 * adjust wait_time so that we'll wake up as soon as it can be
+		 * started.
+		 *
+		 * Each apply worker can only be restarted once per
+		 * wal_retrieve_retry_interval, so that errors do not cause us to
+		 * repeatedly restart the worker as fast as possible.
+		 */
+		last_sync = slot_db_data->last_sync_time;
+		now = GetCurrentTimestamp();
+		if (last_sync == 0 ||
+			(elapsed = TimestampDifferenceMilliseconds(last_sync, now)) >=
+			wal_retrieve_retry_interval)
+		{
+			slot_db_data->last_sync_time = now;
+			slot_sync_worker_launch_or_reuse(slot_db_data->database,
+											 BOOTSTRAP_SUPERUSERID);
+		}
+		else
+		{
+			*wait_time = Min(*wait_time,
+							 wal_retrieve_retry_interval - elapsed);
+		}
+	}
+
+	/* Switch back to original memory context. */
+	MemoryContextSwitchTo(oldctx);
+	/* Clean the temporary memory. */
+	MemoryContextDelete(tmpctx);
+
+	walrcv_disconnect(wrconn);
+}
+
+static void
+ApplyLauncherStartSubs(long *wait_time)
+{
+	List	   *sublist;
+	ListCell   *lc;
+	MemoryContext subctx;
+	MemoryContext oldctx;
+
+	/* Use temporary context to avoid leaking memory across cycles. */
+	subctx = AllocSetContextCreate(TopMemoryContext,
+								   "Logical Replication Launcher sublist",
+								   ALLOCSET_DEFAULT_SIZES);
+	oldctx = MemoryContextSwitchTo(subctx);
+
+	/* Start any missing workers for enabled subscriptions. */
+	sublist = get_subscription_list();
+	foreach(lc, sublist)
+	{
+		Subscription *sub = (Subscription *) lfirst(lc);
+		LogicalRepWorker *w;
+		TimestampTz last_start;
+		TimestampTz now;
+		long		elapsed;
+
+		if (!sub->enabled)
+			continue;
+
+		LWLockAcquire(LogicalRepWorkerLock, LW_SHARED);
+		w = logicalrep_worker_find(sub->oid, InvalidOid, false);
+		LWLockRelease(LogicalRepWorkerLock);
+
+		if (w != NULL)
+			continue;			/* worker is running already */
+
+		/*
+		 * If the worker is eligible to start now, launch it.  Otherwise,
+		 * adjust wait_time so that we'll wake up as soon as it can be
+		 * started.
+		 *
+		 * Each subscription's apply worker can only be restarted once per
+		 * wal_retrieve_retry_interval, so that errors do not cause us to
+		 * repeatedly restart the worker as fast as possible.  In cases where
+		 * a restart is expected (e.g., subscription parameter changes),
+		 * another process should remove the last-start entry for the
+		 * subscription so that the worker can be restarted without waiting
+		 * for wal_retrieve_retry_interval to elapse.
+		 */
+		last_start = ApplyLauncherGetWorkerStartTime(sub->oid);
+		now = GetCurrentTimestamp();
+		if (last_start == 0 ||
+			(elapsed = TimestampDifferenceMilliseconds(last_start, now)) >= wal_retrieve_retry_interval)
+		{
+			ApplyLauncherSetWorkerStartTime(sub->oid, now);
+			logicalrep_worker_launch(WORKERTYPE_APPLY,
+									 sub->dbid, sub->oid, sub->name,
+									 sub->owner, InvalidOid,
+									 DSM_HANDLE_INVALID);
+		}
+		else
+		{
+			*wait_time = Min(*wait_time,
+							 wal_retrieve_retry_interval - elapsed);
+		}
+	}
+
+	/* Switch back to original memory context. */
+	MemoryContextSwitchTo(oldctx);
+	/* Clean the temporary memory. */
+	MemoryContextDelete(subctx);
+}
+
 /*
  * Main loop for the apply launcher process.
  */
@@ -1139,79 +1899,20 @@ ApplyLauncherMain(Datum main_arg)
 	 */
 	BackgroundWorkerInitializeConnection(NULL, NULL, 0);
 
+	load_file("libpqwalreceiver", false);
+
 	/* Enter main loop */
 	for (;;)
 	{
 		int			rc;
-		List	   *sublist;
-		ListCell   *lc;
-		MemoryContext subctx;
-		MemoryContext oldctx;
 		long		wait_time = DEFAULT_NAPTIME_PER_CYCLE;
 
 		CHECK_FOR_INTERRUPTS();
 
-		/* Use temporary context to avoid leaking memory across cycles. */
-		subctx = AllocSetContextCreate(TopMemoryContext,
-									   "Logical Replication Launcher sublist",
-									   ALLOCSET_DEFAULT_SIZES);
-		oldctx = MemoryContextSwitchTo(subctx);
-
-		/* Start any missing workers for enabled subscriptions. */
-		sublist = get_subscription_list();
-		foreach(lc, sublist)
-		{
-			Subscription *sub = (Subscription *) lfirst(lc);
-			LogicalRepWorker *w;
-			TimestampTz last_start;
-			TimestampTz now;
-			long		elapsed;
-
-			if (!sub->enabled)
-				continue;
-
-			LWLockAcquire(LogicalRepWorkerLock, LW_SHARED);
-			w = logicalrep_worker_find(sub->oid, InvalidOid, false);
-			LWLockRelease(LogicalRepWorkerLock);
-
-			if (w != NULL)
-				continue;		/* worker is running already */
-
-			/*
-			 * If the worker is eligible to start now, launch it.  Otherwise,
-			 * adjust wait_time so that we'll wake up as soon as it can be
-			 * started.
-			 *
-			 * Each subscription's apply worker can only be restarted once per
-			 * wal_retrieve_retry_interval, so that errors do not cause us to
-			 * repeatedly restart the worker as fast as possible.  In cases
-			 * where a restart is expected (e.g., subscription parameter
-			 * changes), another process should remove the last-start entry
-			 * for the subscription so that the worker can be restarted
-			 * without waiting for wal_retrieve_retry_interval to elapse.
-			 */
-			last_start = ApplyLauncherGetWorkerStartTime(sub->oid);
-			now = GetCurrentTimestamp();
-			if (last_start == 0 ||
-				(elapsed = TimestampDifferenceMilliseconds(last_start, now)) >= wal_retrieve_retry_interval)
-			{
-				ApplyLauncherSetWorkerStartTime(sub->oid, now);
-				logicalrep_worker_launch(WORKERTYPE_APPLY,
-										 sub->dbid, sub->oid, sub->name,
-										 sub->owner, InvalidOid,
-										 DSM_HANDLE_INVALID);
-			}
-			else
-			{
-				wait_time = Min(wait_time,
-								wal_retrieve_retry_interval - elapsed);
-			}
-		}
-
-		/* Switch back to original memory context. */
-		MemoryContextSwitchTo(oldctx);
-		/* Clean the temporary memory. */
-		MemoryContextDelete(subctx);
+		if (!RecoveryInProgress())
+			ApplyLauncherStartSubs(&wait_time);
+		else
+			ApplyLauncherStartSlotSync(&wait_time);
 
 		/* Wait for more work. */
 		rc = WaitLatch(MyLatch,
diff --git a/src/backend/replication/logical/logicalfuncs.c b/src/backend/replication/logical/logicalfuncs.c
index 197169d6b0..de318fb29c 100644
--- a/src/backend/replication/logical/logicalfuncs.c
+++ b/src/backend/replication/logical/logicalfuncs.c
@@ -202,6 +202,19 @@ pg_logical_slot_get_changes_guts(FunctionCallInfo fcinfo, bool confirm, bool bin
 
 	ReplicationSlotAcquire(NameStr(*name), true);
 
+	/*
+	 * Do not allow consumption of a "synchronized" slot until the standby
+	 * gets promoted.
+	 */
+	if (RecoveryInProgress() && MyReplicationSlot->data.synced)
+	{
+		ReplicationSlotRelease();
+		ereport(ERROR,
+				(errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
+				 errmsg("operation not permitted on replication slots on "
+						"standby which are synchronized from primary")));
+	}
+
 	PG_TRY();
 	{
 		/* restart at slot's confirmed_flush */
diff --git a/src/backend/replication/logical/meson.build b/src/backend/replication/logical/meson.build
index d48cd4c590..9e52ec421f 100644
--- a/src/backend/replication/logical/meson.build
+++ b/src/backend/replication/logical/meson.build
@@ -11,6 +11,7 @@ backend_sources += files(
   'proto.c',
   'relation.c',
   'reorderbuffer.c',
+  'slotsync.c',
   'snapbuild.c',
   'tablesync.c',
   'worker.c',
diff --git a/src/backend/replication/logical/slotsync.c b/src/backend/replication/logical/slotsync.c
new file mode 100644
index 0000000000..d1a3435753
--- /dev/null
+++ b/src/backend/replication/logical/slotsync.c
@@ -0,0 +1,972 @@
+/*-------------------------------------------------------------------------
+ * slotsync.c
+ *	   PostgreSQL worker for synchronizing slots to a standby from primary
+ *
+ * Copyright (c) 2016-2018, PostgreSQL Global Development Group
+ *
+ * IDENTIFICATION
+ *	  src/backend/replication/logical/slotsync.c
+ *
+ * This file contains the code for slot-sync worker on physical standby that
+ * fetches the logical replication slots information from primary server
+ * (PrimaryConnInfo) and creates the slots on standby and synchronizes them
+ * periodically. It synchronizes only the slots configured in
+ * 'synchronize_slot_names'.
+ *
+ * It also takes care of dropping the slots which were created by it and are
+ * currently not needed to be synchronized.
+ *
+ * It takes a nap of WORKER_DEFAULT_NAPTIME before every next synchronization.
+ * If there is no acitivity observed on primary for sometime, it increases the
+ * naptime to WORKER_INACTIVITY_NAPTIME and as soon as any activity is observed,
+ * it brings back the naptime to default value.
+ *-------------------------------------------------------------------------
+ */
+
+#include "postgres.h"
+
+#include "commands/dbcommands.h"
+#include "pgstat.h"
+#include "postmaster/bgworker.h"
+#include "postmaster/interrupt.h"
+#include "replication/logical.h"
+#include "replication/logicallauncher.h"
+#include "replication/logicalworker.h"
+#include "replication/walreceiver.h"
+#include "replication/worker_internal.h"
+#include "storage/ipc.h"
+#include "storage/procarray.h"
+#include "tcop/tcopprot.h"
+#include "utils/builtins.h"
+#include "utils/guc_hooks.h"
+#include "utils/pg_lsn.h"
+#include "utils/varlena.h"
+
+typedef struct RemoteSlot
+{
+	char	   *name;
+	char	   *plugin;
+	char	   *database;
+	bool		two_phase;
+	bool		conflicting;
+	XLogRecPtr	restart_lsn;
+	XLogRecPtr	confirmed_lsn;
+	TransactionId catalog_xmin;
+
+	/* RS_INVAL_NONE if valid, or the reason of invalidation */
+	ReplicationSlotInvalidationCause invalidated;
+} RemoteSlot;
+
+List	   *synchronize_slot_names_list = NIL;
+
+/* Worker's naptime in case of regular activity on primary */
+#define WORKER_DEFAULT_NAPTIME                   10L	/* 10 ms */
+
+/* Worker's naptime in case of no-activity on primary */
+#define WORKER_INACTIVITY_NAPTIME        10000L /* 10 sec */
+
+/*
+ * Inactivity Threshold Count before increasing naptime of worker.
+ *
+ * If the lsn of slot being monitored did not change for these many times,
+ * then increase naptime of current worker from WORKER_DEFAULT_NAPTIME to
+ * WORKER_INACTIVITY_NAPTIME.
+ */
+#define WORKER_INACTIVITY_THRESHOLD 10
+
+/*
+ * Wait for remote slot to pass localy reserved position.
+ */
+static void
+wait_for_primary_slot_catchup(WalReceiverConn *wrconn, char *slot_name,
+							  XLogRecPtr min_lsn)
+{
+	WalRcvExecResult *res;
+	TupleTableSlot *slot;
+	Oid			slotRow[1] = {LSNOID};
+	StringInfoData cmd;
+	bool		isnull;
+	XLogRecPtr	restart_lsn;
+
+	for (;;)
+	{
+		int			rc;
+
+		CHECK_FOR_INTERRUPTS();
+
+		initStringInfo(&cmd);
+		appendStringInfo(&cmd,
+						 "SELECT restart_lsn"
+						 "  FROM pg_catalog.pg_replication_slots"
+						 " WHERE slot_name = %s",
+						 quote_literal_cstr(slot_name));
+		res = walrcv_exec(wrconn, cmd.data, 1, slotRow);
+		pfree(cmd.data);
+
+		if (res->status != WALRCV_OK_TUPLES)
+			ereport(ERROR,
+					(errmsg("could not fetch slot info for slot \"%s\" from"
+							" primary: %s", slot_name, res->err)));
+
+		slot = MakeSingleTupleTableSlot(res->tupledesc, &TTSOpsMinimalTuple);
+		if (!tuplestore_gettupleslot(res->tuplestore, true, false, slot))
+			ereport(ERROR,
+					(errmsg("slot \"%s\" disapeared from primary",
+							slot_name)));
+
+		restart_lsn = DatumGetLSN(slot_getattr(slot, 1, &isnull));
+		Assert(!isnull);
+
+		ExecClearTuple(slot);
+		walrcv_clear_result(res);
+
+		if (restart_lsn >= min_lsn)
+			break;
+
+		rc = WaitLatch(MyLatch,
+					   WL_LATCH_SET | WL_TIMEOUT | WL_POSTMASTER_DEATH,
+					   wal_retrieve_retry_interval,
+					   WAIT_EVENT_REPL_SLOT_SYNC_MAIN);
+
+		ResetLatch(MyLatch);
+
+		/* emergency bailout if postmaster has died */
+		if (rc & WL_POSTMASTER_DEATH)
+			proc_exit(1);
+	}
+}
+
+/*
+ * Update local slot metadata as per remote_slot's positions
+ */
+static void
+local_slot_update(RemoteSlot *remote_slot)
+{
+	LogicalConfirmReceivedLocation(remote_slot->confirmed_lsn);
+	LogicalIncreaseXminForSlot(remote_slot->confirmed_lsn,
+							   remote_slot->catalog_xmin);
+	LogicalIncreaseRestartDecodingForSlot(remote_slot->confirmed_lsn,
+										  remote_slot->restart_lsn);
+	MyReplicationSlot->data.invalidated = remote_slot->invalidated;
+	ReplicationSlotMarkDirty();
+}
+
+/*
+ * Get list of local logical slot names which are synchronized from
+ * primary and belongs to one of the DBs passed in.
+ */
+static List *
+get_local_synced_slot_names(Oid *dbids)
+{
+	List	   *localSyncedSlots = NIL;
+
+	Assert(LWLockHeldByMe(SlotSyncWorkerLock));
+
+	LWLockAcquire(ReplicationSlotControlLock, LW_SHARED);
+
+	for (int i = 0; i < max_replication_slots; i++)
+	{
+		ReplicationSlot *s = &ReplicationSlotCtl->replication_slots[i];
+
+		/* Check if it is logical synchronized slot */
+		if (s->in_use && SlotIsLogical(s) && s->data.synced)
+		{
+			for (int j = 0; j < MySlotSyncWorker->dbcount; j++)
+			{
+				/*
+				 * Add it to output list if this belongs to one of the
+				 * worker's dbs.
+				 */
+				if (s->data.database == dbids[j])
+				{
+					localSyncedSlots = lappend(localSyncedSlots, s);
+					break;
+				}
+			}
+		}
+	}
+
+	LWLockRelease(ReplicationSlotControlLock);
+
+	return localSyncedSlots;
+}
+
+/*
+ * Helper function to check if local_slot is present in remote_slots list.
+ *
+ * It also checks if logical slot is locally invalidated i.e. invalided on
+ * standby but valid on primary. If found so, it sets locally_invalidated to
+ * true.
+ */
+static bool
+slot_exists_locally(List *remote_slots, ReplicationSlot *local_slot,
+					bool *locally_invalidated)
+{
+	ListCell   *cell;
+
+	foreach(cell, remote_slots)
+	{
+		RemoteSlot *remote_slot = (RemoteSlot *) lfirst(cell);
+
+		if (strcmp(remote_slot->name, NameStr(local_slot->data.name)) == 0)
+		{
+			/*
+			 * if remote slot is marked as non-conflicting (i.e. not
+			 * invalidated) but local slot is marked as invalidated, then set
+			 * the bool.
+			 */
+			if (!remote_slot->conflicting &&
+				SlotIsLogical(local_slot) &&
+				local_slot->data.invalidated != RS_INVAL_NONE)
+				*locally_invalidated = true;
+
+			return true;
+		}
+	}
+
+	return false;
+}
+
+/*
+ * Use slot_name in query.
+ *
+ * Check the dbid of the slot and if the dbid is one of the dbids managed by
+ * current worker, then use this slot-name in query to get the data from
+ * primary. If the slot is not created yet on standby (first time it is being
+ * queried), then too, use this slot in query.
+ */
+static bool
+use_slot_in_query(char *slot_name, Oid *dbids)
+{
+	bool		slot_found = false;
+	bool		relevant_db = false;
+	ReplicationSlot *slot;
+
+	Assert(LWLockHeldByMe(SlotSyncWorkerLock));
+
+	/* Search for the local slot with the same name as slot_name */
+	LWLockAcquire(ReplicationSlotControlLock, LW_SHARED);
+
+	for (int i = 0; i < max_replication_slots; i++)
+	{
+		ReplicationSlot *s = &ReplicationSlotCtl->replication_slots[i];
+
+		if (s->in_use && SlotIsLogical(s) &&
+			(strcmp(NameStr(s->data.name), slot_name) == 0))
+		{
+			slot_found = true;
+			slot = s;
+			break;
+		}
+	}
+
+	/* Check if slot belongs to one of the input dbids */
+	if (slot_found)
+	{
+		for (int j = 0; j < MySlotSyncWorker->dbcount; j++)
+		{
+			if (slot->data.database == dbids[j])
+			{
+				relevant_db = true;
+				break;
+			}
+		}
+	}
+
+	LWLockRelease(ReplicationSlotControlLock);
+
+	/*
+	 * Return TRUE if either slot is not yet created on standby or if it
+	 * belongs to one of the dbs passed in dbids.
+	 */
+	if (!slot_found || relevant_db)
+		return true;
+
+	return false;
+}
+
+
+/*
+ * Compute naptime for MySlotSyncWorker.
+ *
+ * The slot-sync worker takes a nap before it again checks for slots on primary.
+ * The time for each nap is computed here.
+ *
+ * The first slot managed by each worker is chosen for monitoring purpose.
+ * If the lsn of that slot changes during each sync-check time, then the
+ * naptime is kept at regular value of WORKER_DEFAULT_NAPTIME.
+ * When no lsn change is observed for contiguous WORKER_INACTIVITY_THRESHOLD
+ * times, then the naptime is increased to WORKER_INACTIVITY_NAPTIME.
+ * This naptime is brought back to WORKER_DEFAULT_NAPTIME as soon as lsn change
+ * is observed.
+ *
+ * The caller is supposed to ignore return-value of 0. The 0 value is returned
+ * for the slots other that slot being monitored.
+ */
+static long
+compute_naptime(RemoteSlot *remote_slot)
+{
+	if (NameStr(MySlotSyncWorker->monitoring_info.slot_name)[0] == '\0')
+	{
+		/*
+		 * First time, just update the name and lsn and return regular
+		 * naptime. Start comparison from next time onward.
+		 */
+		strcpy(NameStr(MySlotSyncWorker->monitoring_info.slot_name),
+			   remote_slot->name);
+		MySlotSyncWorker->monitoring_info.confirmed_lsn =
+			remote_slot->confirmed_lsn;
+		return WORKER_DEFAULT_NAPTIME;
+	}
+
+	/* If this is the slot being monitored by this worker, compute naptime */
+	if (strcmp(remote_slot->name,
+			   NameStr(MySlotSyncWorker->monitoring_info.slot_name)) == 0)
+	{
+		/*
+		 * if last lsn (monitored one) is same as current lsn (remote one),
+		 * increment inactivity_count.
+		 */
+		if (MySlotSyncWorker->monitoring_info.confirmed_lsn ==
+			remote_slot->confirmed_lsn)
+			MySlotSyncWorker->monitoring_info.inactivity_count++;
+		else
+			MySlotSyncWorker->monitoring_info.inactivity_count = 0;
+
+		MySlotSyncWorker->monitoring_info.confirmed_lsn =
+			remote_slot->confirmed_lsn;
+
+		/* If inactivity_count reaches the threshold, increase naptime */
+		if (MySlotSyncWorker->monitoring_info.inactivity_count >=
+			WORKER_INACTIVITY_THRESHOLD)
+			return WORKER_INACTIVITY_NAPTIME;
+		else
+			return WORKER_DEFAULT_NAPTIME;
+	}
+
+	/* if it is not the slot being monitored, return 0 */
+	return 0;
+}
+
+/*
+ * Get Remote Slot's invalidation cause.
+ *
+ * This gets invalidation cause of remote slot.
+ */
+static ReplicationSlotInvalidationCause
+get_remote_invalidation_cause(WalReceiverConn *wrconn, char *slot_name)
+{
+	WalRcvExecResult *res;
+	Oid			slotRow[1] = {INT2OID};
+	StringInfoData cmd;
+	bool		isnull;
+	TupleTableSlot *slot;
+	ReplicationSlotInvalidationCause cause;
+	MemoryContext oldctx = CurrentMemoryContext;
+
+	/* syscache access needs a transaction env. */
+	StartTransactionCommand();
+
+	/* make things live outside TX context */
+	MemoryContextSwitchTo(oldctx);
+
+	initStringInfo(&cmd);
+	appendStringInfo(&cmd,
+					 "select pg_get_invalidation_cause(%s)",
+					 quote_literal_cstr(slot_name));
+	res = walrcv_exec(wrconn, cmd.data, 1, slotRow);
+	pfree(cmd.data);
+
+	CommitTransactionCommand();
+
+	/* Switch to oldctx we saved */
+	MemoryContextSwitchTo(oldctx);
+
+	if (res->status != WALRCV_OK_TUPLES)
+		ereport(ERROR,
+				(errmsg("could not fetch invalidation cuase for slot \"%s\" from"
+						" primary: %s", slot_name, res->err)));
+
+	slot = MakeSingleTupleTableSlot(res->tupledesc, &TTSOpsMinimalTuple);
+	if (!tuplestore_gettupleslot(res->tuplestore, true, false, slot))
+		ereport(ERROR,
+				(errmsg("slot \"%s\" disapeared from primary",
+						slot_name)));
+
+	cause = DatumGetInt16(slot_getattr(slot, 1, &isnull));
+	Assert(!isnull);
+
+	ExecClearTuple(slot);
+	walrcv_clear_result(res);
+
+	return cause;
+}
+
+/*
+ * Drop obsolete slots
+ *
+ * Drop the slots which no longer need to be synced i.e. these either
+ * do not exist on primary or are no longer part of synchronize_slot_names.
+ *
+ * Also drop the slots which are valid on primary and got invalidated
+ * on standby due to conflict (say required rows removed on primary).
+ * The assumption is, these will get recreated in next sync-cycle and
+ * it is okay to drop and recreate such slots as long as these are not
+ * consumable on standby (which is the case currently).
+ */
+static void
+drop_obsolete_slots(Oid *dbids, List *remote_slot_list)
+{
+	List	   *local_slot_list = NIL;
+	ListCell   *lc_slot;
+
+	Assert(LWLockHeldByMeInMode(SlotSyncWorkerLock, LW_SHARED));
+
+	/*
+	 * Get the list of local slots for dbids managed by this worker, so that
+	 * those not on remote could be dropped.
+	 */
+	local_slot_list = get_local_synced_slot_names(dbids);
+
+	foreach(lc_slot, local_slot_list)
+	{
+		ReplicationSlot *local_slot = (ReplicationSlot *) lfirst(lc_slot);
+		bool		local_exists = false;
+		bool		locally_invalidated = false;
+
+		local_exists = slot_exists_locally(remote_slot_list, local_slot,
+										   &locally_invalidated);
+
+		/*
+		 * Drop the local slot either if it is not in the remote slots list or
+		 * is invalidated while remote slot is still valid.
+		 */
+		if (!local_exists || locally_invalidated)
+		{
+			ReplicationSlotDrop(NameStr(local_slot->data.name), true);
+
+			/* if this slot is being monitored, clean-up the monitoring info */
+			if (strcmp(NameStr(local_slot->data.name),
+					   NameStr(MySlotSyncWorker->monitoring_info.slot_name)) == 0)
+			{
+				MemSet(NameStr(MySlotSyncWorker->monitoring_info.slot_name),
+					   0, NAMEDATALEN);
+				MySlotSyncWorker->monitoring_info.inactivity_count = 0;
+				MySlotSyncWorker->monitoring_info.confirmed_lsn = 0;
+			}
+
+			elog(LOG, "Dropped replication slot \"%s\" ",
+				 NameStr(local_slot->data.name));
+		}
+	}
+}
+
+/*
+ * Construct Slot Query
+ *
+ * It constructs the query using dbids array and synchronize_slot_names_list
+ * in order to get slots information from primary.
+ */
+static bool
+construct_slot_query(StringInfo s, Oid *dbids)
+{
+	Assert(LWLockHeldByMeInMode(SlotSyncWorkerLock, LW_SHARED));
+
+	appendStringInfo(s,
+					 "SELECT slot_name, plugin, confirmed_flush_lsn,"
+					 " restart_lsn, catalog_xmin, two_phase, conflicting, "
+					 " database FROM pg_catalog.pg_replication_slots"
+					 " WHERE database IN ");
+
+	/*
+	 * There is a small gap between CHECK_FOR_INTERRUPTS done last and this
+	 * stage, so there is a chance that synchronize_slot_names has changed
+	 * making dbs assigned to this worker as invalid. In that case, launcher
+	 * will make dbcount=0 and will send SIGINT to this worker. So check
+	 * dbcount before proceeding.
+	 */
+	if (!MySlotSyncWorker->dbcount)
+	{
+		/* return and let the worker handle the interrupts in main loop */
+		return false;
+	}
+
+	appendStringInfoChar(s, '(');
+	for (int i = 0; i < MySlotSyncWorker->dbcount; i++)
+	{
+		char	   *dbname;
+
+		if (i != 0)
+			appendStringInfoChar(s, ',');
+
+		dbname = get_database_name(dbids[i]);
+		appendStringInfo(s, "%s",
+						 quote_literal_cstr(dbname));
+		pfree(dbname);
+	}
+	appendStringInfoChar(s, ')');
+
+	if (strcmp(synchronize_slot_names, "") != 0 &&
+		strcmp(synchronize_slot_names, "*") != 0)
+	{
+		ListCell   *lc;
+		bool		first_slot = true;
+
+
+		foreach(lc, synchronize_slot_names_list)
+		{
+			char	   *slot_name = lfirst(lc);
+
+			if (!use_slot_in_query(slot_name, dbids))
+				continue;
+
+			if (first_slot)
+				appendStringInfoString(s, " AND slot_name IN (");
+			else
+				appendStringInfoChar(s, ',');
+
+			appendStringInfo(s, "%s",
+							 quote_literal_cstr(slot_name));
+			first_slot = false;
+		}
+
+		if (!first_slot)
+			appendStringInfoChar(s, ')');
+	}
+
+	return true;
+}
+
+/*
+ * Synchronize single slot to given position.
+ *
+ * This creates new slot if there is no existing one and updates the
+ * metadata of existing slots as per the data received from primary.
+ */
+static void
+synchronize_one_slot(WalReceiverConn *wrconn, RemoteSlot *remote_slot)
+{
+	bool		found = false;
+
+	/*
+	 * Make sure that concerned WAL is received before syncing slot to target
+	 * lsn received from the primary.
+	 */
+	if (remote_slot->confirmed_lsn > WalRcv->latestWalEnd)
+	{
+		ereport(WARNING,
+				errmsg("skipping sync of slot \"%s\" as the received slot-sync "
+					   "lsn %X/%X is ahead of the standby position %X/%X",
+					   remote_slot->name,
+					   LSN_FORMAT_ARGS(remote_slot->confirmed_lsn),
+					   LSN_FORMAT_ARGS(WalRcv->latestWalEnd)));
+		return;
+	}
+
+	/* Search for the named slot and mark it active if we find it. */
+	LWLockAcquire(ReplicationSlotControlLock, LW_SHARED);
+	for (int i = 0; i < max_replication_slots; i++)
+	{
+		ReplicationSlot *s = &ReplicationSlotCtl->replication_slots[i];
+
+		if (!s->in_use)
+			continue;
+
+		if (strcmp(NameStr(s->data.name), remote_slot->name) == 0)
+		{
+			found = true;
+			break;
+		}
+	}
+	LWLockRelease(ReplicationSlotControlLock);
+
+	StartTransactionCommand();
+
+	/* Already existing slot, acquire */
+	if (found)
+	{
+		ReplicationSlotAcquire(remote_slot->name, true);
+
+		if (remote_slot->confirmed_lsn < MyReplicationSlot->data.confirmed_flush)
+		{
+			ereport(WARNING,
+					errmsg("not synchronizing slot %s; synchronization would move"
+						   " it backward", remote_slot->name));
+
+			ReplicationSlotRelease();
+			CommitTransactionCommand();
+			return;
+		}
+
+		/* update lsns of slot to remote slot's current position */
+		local_slot_update(remote_slot);
+		ReplicationSlotSave();
+	}
+	/* Otherwise create the slot first. */
+	else
+	{
+		TransactionId xmin_horizon = InvalidTransactionId;
+		ReplicationSlot *slot;
+
+		ReplicationSlotCreate(remote_slot->name, true, RS_EPHEMERAL,
+							  remote_slot->two_phase);
+		slot = MyReplicationSlot;
+
+		SpinLockAcquire(&slot->mutex);
+		slot->data.database = get_database_oid(remote_slot->database, false);
+		slot->data.synced = true;
+		namestrcpy(&slot->data.plugin, remote_slot->plugin);
+		SpinLockRelease(&slot->mutex);
+
+		ReplicationSlotReserveWal();
+
+		LWLockAcquire(ProcArrayLock, LW_EXCLUSIVE);
+		xmin_horizon = GetOldestSafeDecodingTransactionId(true);
+		slot->effective_catalog_xmin = xmin_horizon;
+		slot->data.catalog_xmin = xmin_horizon;
+		ReplicationSlotsComputeRequiredXmin(true);
+		LWLockRelease(ProcArrayLock);
+
+		if (remote_slot->confirmed_lsn < MyReplicationSlot->data.restart_lsn)
+		{
+			ereport(LOG,
+					errmsg("waiting for remote slot \"%s\" LSN (%X/%X) to pass "
+						   "local slot LSN (%X/%X)", remote_slot->name,
+						   LSN_FORMAT_ARGS(remote_slot->confirmed_lsn),
+						   LSN_FORMAT_ARGS(MyReplicationSlot->data.restart_lsn)));
+
+			wait_for_primary_slot_catchup(wrconn, remote_slot->name,
+										  MyReplicationSlot->data.restart_lsn);
+		}
+
+
+		/* update lsns of slot to remote slot's current position */
+		local_slot_update(remote_slot);
+		ReplicationSlotPersist();
+	}
+
+	ReplicationSlotRelease();
+	CommitTransactionCommand();
+}
+
+/*
+ * Synchronize slots.
+ *
+ * It looks into dbids array maintained in dsa and gets the logical slots info
+ * from primary for the slots configured in synchronize_slot_names and belonging
+ * to concerned dbids. It then updates the slots locally as per the data
+ * received from primary. It creates the slots if not present on standby.
+ *
+ * It returns naptime for the next sync-cycle.
+ */
+static long
+synchronize_slots(dsa_area *dsa)
+{
+#define SLOTSYNC_COLUMN_COUNT 8
+	Oid			slotRow[SLOTSYNC_COLUMN_COUNT] = {TEXTOID, TEXTOID, LSNOID,
+	LSNOID, XIDOID, BOOLOID, BOOLOID, TEXTOID};
+
+	WalRcvExecResult *res;
+	WalReceiverConn *wrconn = NULL;
+	TupleTableSlot *slot;
+	StringInfoData s;
+	List	   *remote_slot_list = NIL;
+	char	   *database;
+	char	   *err;
+	MemoryContext oldctx = CurrentMemoryContext;
+	long		naptime = WORKER_DEFAULT_NAPTIME;
+	long		value;
+	Oid		   *dbids;
+	ListCell   *cell;
+
+	if (strcmp(synchronize_slot_names, "") == 0)
+		return naptime;
+
+	/* WalRcvData is not set or primary_slot_name is not set yet */
+	if (!WalRcv || WalRcv->slotname[0] == '\0')
+		return naptime;
+
+	/* WALs not received yet */
+	if (XLogRecPtrIsInvalid(WalRcv->latestWalEnd))
+		return naptime;
+
+	/* syscache access needs a transaction env. */
+	StartTransactionCommand();
+
+	/* make things live outside TX context */
+	MemoryContextSwitchTo(oldctx);
+
+	database = get_database_name(MyDatabaseId);
+
+	/* Make connection to primary */
+	initStringInfo(&s);
+	appendStringInfo(&s, "%s dbname=%s", PrimaryConnInfo, database);
+	pfree(database);
+
+	wrconn = walrcv_connect(s.data, true, false, "slot_sync", &err);
+	if (wrconn == NULL)
+		ereport(ERROR,
+				(errmsg("could not connect to the primary server: %s", err)));
+
+
+	/* Get dbids from dsa */
+
+	LWLockAcquire(SlotSyncWorkerLock, LW_SHARED);
+	SpinLockAcquire(&MySlotSyncWorker->mutex);
+
+	dbids = (Oid *) dsa_get_address(dsa, MySlotSyncWorker->dbids_dp);
+
+	/* Construct query to get slots info from primary */
+	resetStringInfo(&s);
+	if (!construct_slot_query(&s, dbids))
+	{
+		pfree(s.data);
+		CommitTransactionCommand();
+		SpinLockRelease(&MySlotSyncWorker->mutex);
+		LWLockRelease(SlotSyncWorkerLock);
+		return naptime;
+	}
+
+	elog(DEBUG2, "slot-sync worker%d's query:%s \n", MySlotSyncWorker->slot,
+		 s.data);
+
+	/* Execute the query */
+	res = walrcv_exec(wrconn, s.data, SLOTSYNC_COLUMN_COUNT, slotRow);
+	pfree(s.data);
+
+	if (res->status != WALRCV_OK_TUPLES)
+		ereport(ERROR,
+				(errmsg("could not fetch slot info from primary: %s",
+						res->err)));
+
+	CommitTransactionCommand();
+
+	/* Switch to oldctx we saved */
+	MemoryContextSwitchTo(oldctx);
+
+	/* Construct the remote_slot tuple and synchronize each slot locally */
+	slot = MakeSingleTupleTableSlot(res->tupledesc, &TTSOpsMinimalTuple);
+	while (tuplestore_gettupleslot(res->tuplestore, true, false, slot))
+	{
+		bool		isnull;
+		RemoteSlot *remote_slot = palloc0(sizeof(RemoteSlot));
+
+		remote_slot->name = TextDatumGetCString(slot_getattr(slot, 1, &isnull));
+		Assert(!isnull);
+
+		remote_slot->plugin = TextDatumGetCString(slot_getattr(slot, 2, &isnull));
+		Assert(!isnull);
+
+		/*
+		 * It is possible to get null values for lsns and xmin if slot is
+		 * invalidated on primary, so handle accordingly.
+		 */
+		remote_slot->confirmed_lsn = DatumGetLSN(slot_getattr(slot, 3, &isnull));
+		if (isnull)
+			remote_slot->confirmed_lsn = InvalidXLogRecPtr;
+
+		remote_slot->restart_lsn = DatumGetLSN(slot_getattr(slot, 4, &isnull));
+		if (isnull)
+			remote_slot->restart_lsn = InvalidXLogRecPtr;
+
+		remote_slot->catalog_xmin = DatumGetTransactionId(slot_getattr(slot,
+																	   5, &isnull));
+		if (isnull)
+			remote_slot->catalog_xmin = InvalidTransactionId;
+
+		remote_slot->two_phase = DatumGetBool(slot_getattr(slot, 6, &isnull));
+		Assert(!isnull);
+
+		remote_slot->conflicting = DatumGetBool(slot_getattr(slot, 7, &isnull));
+		Assert(!isnull);
+
+		remote_slot->database = TextDatumGetCString(slot_getattr(slot,
+																 8, &isnull));
+		Assert(!isnull);
+
+		if (remote_slot->conflicting)
+			remote_slot->invalidated = get_remote_invalidation_cause(wrconn,
+																	 remote_slot->name);
+		else
+			remote_slot->invalidated = RS_INVAL_NONE;
+
+		/* Create list of remote slot names to be used by drop_obsolete_slots */
+		remote_slot_list = lappend(remote_slot_list, remote_slot);
+
+		synchronize_one_slot(wrconn, remote_slot);
+
+		/*
+		 * Update naptime in case of non-zero value returned. The zero value
+		 * is returned if remote_slot is not the one being monitored.
+		 */
+		value = compute_naptime(remote_slot);
+		if (value)
+			naptime = value;
+
+		ExecClearTuple(slot);
+	}
+
+	/*
+	 * Drop local slots which no longer need to be synced i.e. these either do
+	 * not exist on primary or are no longer part of synchronize_slot_names.
+	 */
+	drop_obsolete_slots(dbids, remote_slot_list);
+
+	SpinLockRelease(&MySlotSyncWorker->mutex);
+	LWLockRelease(SlotSyncWorkerLock);
+
+	/* We are done, free remot_slot_list elements */
+	foreach(cell, remote_slot_list)
+	{
+		RemoteSlot *remote_slot = (RemoteSlot *) lfirst(cell);
+
+		pfree(remote_slot);
+	}
+
+	walrcv_clear_result(res);
+
+	walrcv_disconnect(wrconn);
+
+	return naptime;
+}
+
+/*
+ * Initialize the list from raw synchronize_slot_names and cache it, in order
+ * to avoid parsing it repeatedly. Done at slot-sync worker startup and after
+ * each SIGHUP.
+ */
+static void
+SlotSyncInitSlotNamesLst()
+{
+	char	   *rawname;
+
+	if (strcmp(synchronize_slot_names, "") != 0 &&
+		strcmp(synchronize_slot_names, "*") != 0)
+	{
+		rawname = pstrdup(synchronize_slot_names);
+		SplitIdentifierString(rawname, ',', &synchronize_slot_names_list);
+	}
+}
+
+/*
+ * Interrupt handler for main loop of slot-sync worker.
+ */
+static void
+ProcessSlotSyncInterrupts()
+{
+	CHECK_FOR_INTERRUPTS();
+
+	if (ShutdownRequestPending)
+	{
+		elog(LOG, "replication slot-sync worker %d is shutting"
+			 " down on receiving SIGINT", MySlotSyncWorker->slot);
+
+		proc_exit(0);
+	}
+
+	if (ConfigReloadPending)
+	{
+		ConfigReloadPending = false;
+		ProcessConfigFile(PGC_SIGHUP);
+		SlotSyncInitSlotNamesLst();
+	}
+}
+
+/*
+ * Detach the worker from DSM and update 'proc' and 'in_use'.
+ * Logical replication launcher will come to know using these
+ * that the worker has shutdown.
+ */
+static void
+slotsync_worker_detach(int code, Datum arg)
+{
+	dsa_detach((dsa_area *) DatumGetPointer(arg));
+	LWLockAcquire(SlotSyncWorkerLock, LW_EXCLUSIVE);
+	MySlotSyncWorker->in_use = false;
+	MySlotSyncWorker->proc = NULL;
+	LWLockRelease(SlotSyncWorkerLock);
+}
+
+/*
+ * The main loop of our worker process.
+ */
+void
+ReplSlotSyncMain(Datum main_arg)
+{
+	int			worker_slot = DatumGetInt32(main_arg);
+	dsa_handle	handle;
+	dsa_area   *dsa;
+
+	/* Setup signal handling */
+	pqsignal(SIGHUP, SignalHandlerForConfigReload);
+	pqsignal(SIGINT, SignalHandlerForShutdownRequest);
+	BackgroundWorkerUnblockSignals();
+
+	/*
+	 * Attach to the dynamic shared memory segment for the slot-sync worker
+	 * and find its table of contents.
+	 */
+	memcpy(&handle, MyBgworkerEntry->bgw_extra, sizeof(dsa_handle));
+	dsa = dsa_attach(handle);
+	if (!dsa)
+		ereport(ERROR,
+				(errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
+				 errmsg("could not map dynamic shared memory "
+						"segment for slot-sync worker")));
+
+
+	/* Primary initialization is complete. Now, attach to our slot. */
+	slotsync_worker_attach(worker_slot);
+
+	before_shmem_exit(slotsync_worker_detach, PointerGetDatum(dsa));
+
+	/* Load the libpq-specific functions */
+	load_file("libpqwalreceiver", false);
+
+	/* Connect to our database. */
+	BackgroundWorkerInitializeConnectionByOid(MySlotSyncWorker->dbid,
+											  MySlotSyncWorker->userid,
+											  0);
+
+	StartTransactionCommand();
+	elog(LOG, "replication slot-sync worker %d "
+		 "started managing database \"%s\" (dbid: %d) ",
+		 worker_slot, get_database_name(MySlotSyncWorker->dbid),
+		 MySlotSyncWorker->dbid);
+	CommitTransactionCommand();
+
+	SlotSyncInitSlotNamesLst();
+
+	/* Main wait loop. */
+	for (;;)
+	{
+		int			rc;
+		long		naptime;
+
+		ProcessSlotSyncInterrupts();
+
+		if (!RecoveryInProgress())
+			proc_exit(0);
+
+		naptime = synchronize_slots(dsa);
+
+		rc = WaitLatch(MyLatch,
+					   WL_LATCH_SET | WL_TIMEOUT | WL_POSTMASTER_DEATH,
+					   naptime,
+					   WAIT_EVENT_REPL_SLOT_SYNC_MAIN);
+
+		ResetLatch(MyLatch);
+
+		/* emergency bailout if postmaster has died */
+		if (rc & WL_POSTMASTER_DEATH)
+			proc_exit(1);
+	}
+
+	/*
+	 * The slot-sync worker must not get here because it will only stop when
+	 * it receives a SIGINT from the logical replication launcher, or when
+	 * there is an error. None of these cases will allow the code to reach
+	 * here.
+	 */
+	Assert(false);
+}
diff --git a/src/backend/replication/logical/tablesync.c b/src/backend/replication/logical/tablesync.c
index e2cee92cf2..36b5ca0898 100644
--- a/src/backend/replication/logical/tablesync.c
+++ b/src/backend/replication/logical/tablesync.c
@@ -100,6 +100,7 @@
 #include "catalog/pg_subscription_rel.h"
 #include "catalog/pg_type.h"
 #include "commands/copy.h"
+#include "commands/subscriptioncmds.h"
 #include "miscadmin.h"
 #include "nodes/makefuncs.h"
 #include "parser/parse_relation.h"
diff --git a/src/backend/replication/repl_gram.y b/src/backend/replication/repl_gram.y
index 0c874e33cf..2b00bf845c 100644
--- a/src/backend/replication/repl_gram.y
+++ b/src/backend/replication/repl_gram.y
@@ -76,11 +76,12 @@ Node *replication_parse_result;
 %token K_EXPORT_SNAPSHOT
 %token K_NOEXPORT_SNAPSHOT
 %token K_USE_SNAPSHOT
+%token K_LIST_DBID_FOR_LOGICAL_SLOTS
 
 %type <node>	command
 %type <node>	base_backup start_replication start_logical_replication
 				create_replication_slot drop_replication_slot identify_system
-				read_replication_slot timeline_history show
+				read_replication_slot timeline_history show list_dbid_for_logical_slots
 %type <list>	generic_option_list
 %type <defelt>	generic_option
 %type <uintval>	opt_timeline
@@ -91,6 +92,7 @@ Node *replication_parse_result;
 %type <boolval>	opt_temporary
 %type <list>	create_slot_options create_slot_legacy_opt_list
 %type <defelt>	create_slot_legacy_opt
+%type <list>	slot_name_list slot_name_list_opt
 
 %%
 
@@ -114,6 +116,7 @@ command:
 			| read_replication_slot
 			| timeline_history
 			| show
+			| list_dbid_for_logical_slots
 			;
 
 /*
@@ -126,6 +129,33 @@ identify_system:
 				}
 			;
 
+slot_name_list:
+			IDENT
+				{
+					$$ = list_make1($1);
+				}
+			| slot_name_list ',' IDENT
+				{
+					$$ = lappend($1, $3);
+				}
+
+slot_name_list_opt:
+			slot_name_list			{ $$ = $1; }
+			| /* EMPTY */			{ $$ = NIL; }
+		;
+
+/*
+ * LIST_DBID_FOR_LOGICAL_SLOTS
+ */
+list_dbid_for_logical_slots:
+			K_LIST_DBID_FOR_LOGICAL_SLOTS slot_name_list_opt
+				{
+					ListDBForLogicalSlotsCmd *cmd = makeNode(ListDBForLogicalSlotsCmd);
+					cmd->slot_names = $2;
+					$$ = (Node *) cmd;
+				}
+			;
+
 /*
  * READ_REPLICATION_SLOT %s
  */
diff --git a/src/backend/replication/repl_scanner.l b/src/backend/replication/repl_scanner.l
index 1cc7fb858c..d4ecce6a47 100644
--- a/src/backend/replication/repl_scanner.l
+++ b/src/backend/replication/repl_scanner.l
@@ -128,6 +128,7 @@ DROP_REPLICATION_SLOT		{ return K_DROP_REPLICATION_SLOT; }
 TIMELINE_HISTORY	{ return K_TIMELINE_HISTORY; }
 PHYSICAL			{ return K_PHYSICAL; }
 RESERVE_WAL			{ return K_RESERVE_WAL; }
+LIST_DBID_FOR_LOGICAL_SLOTS	{ return K_LIST_DBID_FOR_LOGICAL_SLOTS; }
 LOGICAL				{ return K_LOGICAL; }
 SLOT				{ return K_SLOT; }
 TEMPORARY			{ return K_TEMPORARY; }
@@ -304,6 +305,7 @@ replication_scanner_is_replication_command(void)
 		case K_READ_REPLICATION_SLOT:
 		case K_TIMELINE_HISTORY:
 		case K_SHOW:
+		case K_LIST_DBID_FOR_LOGICAL_SLOTS:
 			/* Yes; push back the first token so we can parse later. */
 			repl_pushed_back_token = first_token;
 			return true;
diff --git a/src/backend/replication/slot.c b/src/backend/replication/slot.c
index ad9d6af6fb..09f635e2ea 100644
--- a/src/backend/replication/slot.c
+++ b/src/backend/replication/slot.c
@@ -92,7 +92,7 @@ typedef struct ReplicationSlotOnDisk
 	sizeof(ReplicationSlotOnDisk) - ReplicationSlotOnDiskConstantSize
 
 #define SLOT_MAGIC		0x1051CA1	/* format identifier */
-#define SLOT_VERSION	3		/* version for new files */
+#define SLOT_VERSION	4		/* version for new files */
 
 /* Control array for replication slot management */
 ReplicationSlotCtlData *ReplicationSlotCtl = NULL;
@@ -319,6 +319,7 @@ ReplicationSlotCreate(const char *name, bool db_specific,
 	slot->data.persistency = persistency;
 	slot->data.two_phase = two_phase;
 	slot->data.two_phase_at = InvalidXLogRecPtr;
+	slot->data.synced = false;
 
 	/* and then data only present in shared memory */
 	slot->just_dirtied = false;
diff --git a/src/backend/replication/slotfuncs.c b/src/backend/replication/slotfuncs.c
index 6035cf4816..d1d4593fdb 100644
--- a/src/backend/replication/slotfuncs.c
+++ b/src/backend/replication/slotfuncs.c
@@ -225,6 +225,34 @@ pg_drop_replication_slot(PG_FUNCTION_ARGS)
 	PG_RETURN_VOID();
 }
 
+/*
+ * SQL function for getting invalidation cause of a slot.
+ *
+ * Returns ReplicationSlotInvalidationCause enum value for valid slot_name;
+ * returns NULL if slot with given name is not found.
+ */
+Datum
+pg_get_invalidation_cause(PG_FUNCTION_ARGS)
+{
+	Name		name = PG_GETARG_NAME(0);
+	ReplicationSlotInvalidationCause cause;
+	int			slotno;
+
+	LWLockAcquire(ReplicationSlotControlLock, LW_SHARED);
+	for (slotno = 0; slotno < max_replication_slots; slotno++)
+	{
+		ReplicationSlot *s = &ReplicationSlotCtl->replication_slots[slotno];
+		if (strcmp(NameStr(s->data.name), NameStr(*name)) == 0)
+		{
+			cause = s->data.invalidated;
+			PG_RETURN_INT16(cause);
+		}
+	}
+	LWLockRelease(ReplicationSlotControlLock);
+
+	PG_RETURN_NULL();
+}
+
 /*
  * pg_get_replication_slots - SQL SRF showing all replication slots
  * that currently exist on the database cluster.
diff --git a/src/backend/replication/walsender.c b/src/backend/replication/walsender.c
old mode 100644
new mode 100755
index 0e103ec162..ca4b41f6ac
--- a/src/backend/replication/walsender.c
+++ b/src/backend/replication/walsender.c
@@ -473,6 +473,103 @@ IdentifySystem(void)
 	end_tup_output(tstate);
 }
 
+static int
+pg_qsort_namecmp(const void *a, const void *b)
+{
+	return strncmp(NameStr(*(Name) a), NameStr(*(Name) b), NAMEDATALEN);
+}
+
+/*
+ * Handle the LIST_SLOT_DATABASE_OIDS command.
+ */
+static void
+ListSlotDatabaseOIDs(ListDBForLogicalSlotsCmd *cmd)
+{
+	DestReceiver *dest;
+	TupOutputState *tstate;
+	TupleDesc tupdesc;
+	NameData *slot_names = NULL;
+	int numslot_names;
+	List *database_oids_list = NIL;
+
+	numslot_names = list_length(cmd->slot_names);
+	if (numslot_names)
+	{
+		ListCell *lc;
+		int i = 0;
+
+		slot_names = palloc(numslot_names * sizeof(NameData));
+		foreach (lc, cmd->slot_names)
+		{
+			char *slot_name = lfirst(lc);
+
+			ReplicationSlotValidateName(slot_name, ERROR);
+			namestrcpy(&slot_names[i++], slot_name);
+		}
+
+		qsort(slot_names, numslot_names, sizeof(NameData), pg_qsort_namecmp);
+	}
+
+	dest = CreateDestReceiver(DestRemoteSimple);
+
+	/* need a tuple descriptor representing a single column */
+	tupdesc = CreateTemplateTupleDesc(1);
+	TupleDescInitBuiltinEntry(tupdesc, (AttrNumber)1, "database_oid",
+							  INT8OID, -1, 0);
+
+	/* prepare for projection of tuples */
+	tstate = begin_tup_output_tupdesc(dest, tupdesc, &TTSOpsVirtual);
+
+	LWLockAcquire(ReplicationSlotControlLock, LW_SHARED);
+	for (int slotno = 0; slotno < max_replication_slots; slotno++)
+	{
+		ReplicationSlot *slot = &ReplicationSlotCtl->replication_slots[slotno];
+		Oid datoid; /* Variable to store the database OID for each slot */
+		Datum values[1];
+		bool nulls[1];
+
+		if (!slot->in_use)
+			continue;
+
+		SpinLockAcquire(&slot->mutex);
+
+		datoid = slot->data.database;
+
+		SpinLockRelease(&slot->mutex);
+
+		/*
+		 * If slot names were provided and the current slot name is not in the
+		 * list, skip it.
+		 */
+		if (numslot_names &&
+			!bsearch((void *)&slot->data.name, (void *)slot_names,
+				numslot_names, sizeof(NameData), pg_qsort_namecmp))
+			continue;
+
+		/*
+		 * Check if the database OID is already in the list, and if so, skip
+		 * this slot.
+		 */
+		if ((OidIsValid(datoid) && list_member_oid(database_oids_list, datoid)))
+			continue;
+
+		/* Add the database OID to the list */
+		database_oids_list = lappend_oid(database_oids_list, datoid);
+
+		values[0] = Int64GetDatum(datoid);
+		nulls[0] = (datoid == InvalidOid);
+
+		/* send it to dest */
+		do_tup_output(tstate, values, nulls);
+	}
+	LWLockRelease(ReplicationSlotControlLock);
+
+	/* Clean up the list */
+	list_free(database_oids_list);
+
+	end_tup_output(tstate);
+}
+
 /* Handle READ_REPLICATION_SLOT command */
 static void
 ReadReplicationSlot(ReadReplicationSlotCmd *cmd)
@@ -1823,6 +1920,13 @@ exec_replication_command(const char *cmd_string)
 			EndReplicationCommand(cmdtag);
 			break;
 
+		case T_ListDBForLogicalSlotsCmd:
+			cmdtag = "LIST_DBID_FOR_LOGICAL_SLOTS";
+			set_ps_display(cmdtag);
+			ListSlotDatabaseOIDs((ListDBForLogicalSlotsCmd *) cmd_node);
+			EndReplicationCommand(cmdtag);
+			break;
+
 		case T_StartReplicationCmd:
 			{
 				StartReplicationCmd *cmd = (StartReplicationCmd *) cmd_node;
diff --git a/src/backend/storage/lmgr/lwlock.c b/src/backend/storage/lmgr/lwlock.c
index 315a78cda9..fd9e73a49b 100644
--- a/src/backend/storage/lmgr/lwlock.c
+++ b/src/backend/storage/lmgr/lwlock.c
@@ -190,6 +190,8 @@ static const char *const BuiltinTrancheNames[] = {
 	"LogicalRepLauncherDSA",
 	/* LWTRANCHE_LAUNCHER_HASH: */
 	"LogicalRepLauncherHash",
+	/* LWTRANCHE_SLOTSYNC_DSA: */
+	"SlotSyncWorkerDSA",
 };
 
 StaticAssertDecl(lengthof(BuiltinTrancheNames) ==
diff --git a/src/backend/storage/lmgr/lwlocknames.txt b/src/backend/storage/lmgr/lwlocknames.txt
index f72f2906ce..e62a3f1bc0 100644
--- a/src/backend/storage/lmgr/lwlocknames.txt
+++ b/src/backend/storage/lmgr/lwlocknames.txt
@@ -54,3 +54,4 @@ XactTruncationLock					44
 WrapLimitsVacuumLock				46
 NotifyQueueTailLock					47
 WaitEventExtensionLock				48
+SlotSyncWorkerLock					49
diff --git a/src/backend/utils/activity/wait_event_names.txt b/src/backend/utils/activity/wait_event_names.txt
index 9c5fdeb3ca..0d1fcb14dc 100644
--- a/src/backend/utils/activity/wait_event_names.txt
+++ b/src/backend/utils/activity/wait_event_names.txt
@@ -53,6 +53,7 @@ LOGICAL_APPLY_MAIN	"Waiting in main loop of logical replication apply process."
 LOGICAL_LAUNCHER_MAIN	"Waiting in main loop of logical replication launcher process."
 LOGICAL_PARALLEL_APPLY_MAIN	"Waiting in main loop of logical replication parallel apply process."
 RECOVERY_WAL_STREAM	"Waiting in main loop of startup process for WAL to arrive, during streaming recovery."
+REPL_SLOT_SYNC_MAIN	"Waiting in main loop of worker for synchronizing slots to a standby from primary."
 SYSLOGGER_MAIN	"Waiting in main loop of syslogger process."
 WAL_RECEIVER_MAIN	"Waiting in main loop of WAL receiver process."
 WAL_SENDER_MAIN	"Waiting in main loop of WAL sender process."
diff --git a/src/backend/utils/misc/guc_tables.c b/src/backend/utils/misc/guc_tables.c
index 864f80f328..1d94e3f763 100644
--- a/src/backend/utils/misc/guc_tables.c
+++ b/src/backend/utils/misc/guc_tables.c
@@ -64,8 +64,11 @@
 #include "postmaster/syslogger.h"
 #include "postmaster/walwriter.h"
 #include "replication/logicallauncher.h"
+#include "replication/reorderbuffer.h"
 #include "replication/slot.h"
 #include "replication/syncrep.h"
+#include "replication/walreceiver.h"
+#include "replication/walsender.h"
 #include "storage/bufmgr.h"
 #include "storage/large_object.h"
 #include "storage/pg_shmem.h"
@@ -3497,6 +3500,19 @@ struct config_int ConfigureNamesInt[] =
 		NULL, NULL, NULL
 	},
 
+	{
+		{"max_slotsync_workers",
+			PGC_POSTMASTER,
+			REPLICATION_STANDBY,
+			gettext_noop("Maximum number of slot synchronization workers "
+						 "on a standby."),
+			NULL,
+		},
+		&max_slotsync_workers,
+		2, 0, MAX_SLOT_SYNC_WORKER_LIMIT,
+		NULL, NULL, NULL
+	},
+
 	/* End-of-list marker */
 	{
 		{NULL, 0, 0, NULL, NULL}, NULL, 0, 0, 0, NULL, NULL, NULL
diff --git a/src/backend/utils/misc/postgresql.conf.sample b/src/backend/utils/misc/postgresql.conf.sample
index f1c87bd2d6..d01406452e 100644
--- a/src/backend/utils/misc/postgresql.conf.sample
+++ b/src/backend/utils/misc/postgresql.conf.sample
@@ -357,6 +357,7 @@
 #recovery_min_apply_delay = 0		# minimum delay for applying changes during recovery
 #synchronize_slot_names = ''	# replication slot names to synchronize from
 					# primary to streaming replication standby server
+#max_slotsync_workers = 2		# max number of slot synchronization workers on a standby
 
 # - Subscribers -
 
diff --git a/src/include/catalog/pg_proc.dat b/src/include/catalog/pg_proc.dat
index 9805bc6118..612dc5e0b5 100644
--- a/src/include/catalog/pg_proc.dat
+++ b/src/include/catalog/pg_proc.dat
@@ -11074,6 +11074,10 @@
   proname => 'pg_drop_replication_slot', provolatile => 'v', proparallel => 'u',
   prorettype => 'void', proargtypes => 'name',
   prosrc => 'pg_drop_replication_slot' },
+{ oid => '6312', descr => 'get invalidate cause of a replication slot',
+  proname => 'pg_get_invalidation_cause', provolatile => 's', proisstrict => 't',
+  prorettype => 'int2', proargtypes => 'name',
+  prosrc => 'pg_get_invalidation_cause' },
 { oid => '3781',
   descr => 'information about replication slots currently in use',
   proname => 'pg_get_replication_slots', prorows => '10', proisstrict => 'f',
diff --git a/src/include/commands/subscriptioncmds.h b/src/include/commands/subscriptioncmds.h
index 214dc6c29e..75b4b2040d 100644
--- a/src/include/commands/subscriptioncmds.h
+++ b/src/include/commands/subscriptioncmds.h
@@ -17,6 +17,7 @@
 
 #include "catalog/objectaddress.h"
 #include "parser/parse_node.h"
+#include "replication/walreceiver.h"
 
 extern ObjectAddress CreateSubscription(ParseState *pstate, CreateSubscriptionStmt *stmt,
 										bool isTopLevel);
@@ -28,4 +29,7 @@ extern void AlterSubscriptionOwner_oid(Oid subid, Oid newOwnerId);
 
 extern char defGetStreamingMode(DefElem *def);
 
+extern void ReplicationSlotDropAtPubNode(WalReceiverConn *wrconn,
+										char *slotname, bool missing_ok);
+
 #endif							/* SUBSCRIPTIONCMDS_H */
diff --git a/src/include/nodes/replnodes.h b/src/include/nodes/replnodes.h
index 4321ba8f86..bc9c1baea1 100644
--- a/src/include/nodes/replnodes.h
+++ b/src/include/nodes/replnodes.h
@@ -33,6 +33,15 @@ typedef struct IdentifySystemCmd
 	NodeTag		type;
 } IdentifySystemCmd;
 
+/* -------------------------------
+ *		LIST_DBID_FOR_LOGICAL_SLOTS command
+ * -------------------------------
+ */
+typedef struct ListDBForLogicalSlotsCmd
+{
+	NodeTag		type;
+	List	   *slot_names;
+} ListDBForLogicalSlotsCmd;
 
 /* ----------------------
  *		BASE_BACKUP command
diff --git a/src/include/postmaster/bgworker_internals.h b/src/include/postmaster/bgworker_internals.h
index 4ad63fd9bd..19c5421a55 100644
--- a/src/include/postmaster/bgworker_internals.h
+++ b/src/include/postmaster/bgworker_internals.h
@@ -22,6 +22,7 @@
  * Maximum possible value of parallel workers.
  */
 #define MAX_PARALLEL_WORKER_LIMIT 1024
+#define MAX_SLOT_SYNC_WORKER_LIMIT 50
 
 /*
  * List of background workers, private to postmaster.
diff --git a/src/include/replication/logicallauncher.h b/src/include/replication/logicallauncher.h
index a07c9cb311..891ce08cf1 100644
--- a/src/include/replication/logicallauncher.h
+++ b/src/include/replication/logicallauncher.h
@@ -15,6 +15,8 @@
 extern PGDLLIMPORT int max_logical_replication_workers;
 extern PGDLLIMPORT int max_sync_workers_per_subscription;
 extern PGDLLIMPORT int max_parallel_apply_workers_per_subscription;
+extern PGDLLIMPORT int max_slotsync_workers;
+
 
 extern void ApplyLauncherRegister(void);
 extern void ApplyLauncherMain(Datum main_arg);
@@ -31,4 +33,6 @@ extern bool IsLogicalLauncher(void);
 
 extern pid_t GetLeaderApplyWorkerPid(pid_t pid);
 
+extern PGDLLIMPORT char *PrimaryConnInfo;
+
 #endif							/* LOGICALLAUNCHER_H */
diff --git a/src/include/replication/logicalworker.h b/src/include/replication/logicalworker.h
index bbd71d0b42..e1af29af4a 100644
--- a/src/include/replication/logicalworker.h
+++ b/src/include/replication/logicalworker.h
@@ -19,6 +19,7 @@ extern PGDLLIMPORT volatile sig_atomic_t ParallelApplyMessagePending;
 extern void ApplyWorkerMain(Datum main_arg);
 extern void ParallelApplyWorkerMain(Datum main_arg);
 extern void TablesyncWorkerMain(Datum main_arg);
+extern void ReplSlotSyncMain(Datum main_arg);
 
 extern bool IsLogicalWorker(void);
 extern bool IsLogicalParallelApplyWorker(void);
diff --git a/src/include/replication/slot.h b/src/include/replication/slot.h
index b364acc47a..3628bd0309 100644
--- a/src/include/replication/slot.h
+++ b/src/include/replication/slot.h
@@ -15,7 +15,6 @@
 #include "storage/lwlock.h"
 #include "storage/shmem.h"
 #include "storage/spin.h"
-#include "replication/walreceiver.h"
 
 /*
  * Behaviour of replication slots, upon release or crash.
@@ -111,6 +110,11 @@ typedef struct ReplicationSlotPersistentData
 
 	/* plugin name */
 	NameData	plugin;
+
+	/*
+	 * Is this a slot created by a sync-slot worker?
+	 */
+	bool		synced;
 } ReplicationSlotPersistentData;
 
 /*
@@ -240,7 +244,6 @@ extern ReplicationSlot *SearchNamedReplicationSlot(const char *name, bool need_l
 extern int	ReplicationSlotIndex(ReplicationSlot *slot);
 extern bool ReplicationSlotName(int index, Name name);
 extern void ReplicationSlotNameForTablesync(Oid suboid, Oid relid, char *syncslotname, Size szslot);
-extern void ReplicationSlotDropAtPubNode(WalReceiverConn *wrconn, char *slotname, bool missing_ok);
 
 extern void StartupReplicationSlots(void);
 extern void CheckPointReplicationSlots(void);
diff --git a/src/include/replication/walreceiver.h b/src/include/replication/walreceiver.h
index 281626fa6f..576775f627 100644
--- a/src/include/replication/walreceiver.h
+++ b/src/include/replication/walreceiver.h
@@ -20,6 +20,7 @@
 #include "pgtime.h"
 #include "port/atomics.h"
 #include "replication/logicalproto.h"
+#include "replication/slot.h"
 #include "replication/walsender.h"
 #include "storage/condition_variable.h"
 #include "storage/latch.h"
@@ -191,6 +192,16 @@ typedef struct
 	}			proto;
 } WalRcvStreamOptions;
 
+/*
+ * Slot's DBid related data
+ */
+typedef struct WalRcvRepSlotDbData
+{
+	Oid			database;		/* Slot's DBid received from remote */
+	TimestampTz last_sync_time; /* The last time we tried to launch sync
+								 * worker for above Dbid */
+} WalRcvRepSlotDbData;
+
 struct WalReceiverConn;
 typedef struct WalReceiverConn WalReceiverConn;
 
@@ -280,6 +291,15 @@ typedef void (*walrcv_get_senderinfo_fn) (WalReceiverConn *conn,
 typedef char *(*walrcv_identify_system_fn) (WalReceiverConn *conn,
 											TimeLineID *primary_tli);
 
+/*
+ * walrcv_get_dbinfo_for_logical_slots_fn
+ *
+ * Run LIST_DBID_FOR_LOGICAL_SLOTS on primary server to get the
+ * list of unique DBIDs for logical slots mentioned in 'slots'
+ */
+typedef List *(*walrcv_get_dbinfo_for_logical_slots_fn) (WalReceiverConn *conn,
+													  const char *slots);
+
 /*
  * walrcv_server_version_fn
  *
@@ -393,6 +413,7 @@ typedef struct WalReceiverFunctionsType
 	walrcv_get_conninfo_fn walrcv_get_conninfo;
 	walrcv_get_senderinfo_fn walrcv_get_senderinfo;
 	walrcv_identify_system_fn walrcv_identify_system;
+	walrcv_get_dbinfo_for_logical_slots_fn walrcv_get_dbinfo_for_logical_slots;
 	walrcv_server_version_fn walrcv_server_version;
 	walrcv_readtimelinehistoryfile_fn walrcv_readtimelinehistoryfile;
 	walrcv_startstreaming_fn walrcv_startstreaming;
@@ -417,6 +438,8 @@ extern PGDLLIMPORT WalReceiverFunctionsType *WalReceiverFunctions;
 	WalReceiverFunctions->walrcv_get_senderinfo(conn, sender_host, sender_port)
 #define walrcv_identify_system(conn, primary_tli) \
 	WalReceiverFunctions->walrcv_identify_system(conn, primary_tli)
+#define walrcv_get_dbinfo_for_logical_slots(conn, slots) \
+	WalReceiverFunctions->walrcv_get_dbinfo_for_logical_slots(conn, slots)
 #define walrcv_server_version(conn) \
 	WalReceiverFunctions->walrcv_server_version(conn)
 #define walrcv_readtimelinehistoryfile(conn, tli, filename, content, size) \
diff --git a/src/include/replication/worker_internal.h b/src/include/replication/worker_internal.h
index 8f4bed0958..e401f604c3 100644
--- a/src/include/replication/worker_internal.h
+++ b/src/include/replication/worker_internal.h
@@ -77,7 +77,7 @@ typedef struct LogicalRepWorker
 	 * would be created for each transaction which will be deleted after the
 	 * transaction is finished.
 	 */
-	FileSet    *stream_fileset;
+	struct FileSet *stream_fileset;
 
 	/*
 	 * PID of leader apply worker if this slot is used for a parallel apply
@@ -96,6 +96,51 @@ typedef struct LogicalRepWorker
 	TimestampTz reply_time;
 } LogicalRepWorker;
 
+typedef struct SlotSyncWorker
+{
+	/* Time at which this worker was launched. */
+	TimestampTz launch_time;
+
+	/* Indicates if this slot is used or free. */
+	bool		in_use;
+
+	/* The slot in worker pool to which it is attached */
+	int			slot;
+
+	/* Increased every time the slot is taken by new worker. */
+	uint16		generation;
+
+	/* Pointer to proc array. NULL if not running. */
+	PGPROC	   *proc;
+
+	/* User to use for connection (will be same as owner of subscription). */
+	Oid			userid;
+
+	/* Database id to connect to. */
+	Oid			dbid;
+
+	/* Count of Database ids it manages */
+	uint32		dbcount;
+
+	/* DSA for dbids */
+	dsa_area   *dbids_dsa;
+
+	/* dsa_pointer for database ids it manages */
+	dsa_pointer dbids_dp;
+
+	/* Mutex to access dbids in dsa */
+	slock_t		mutex;
+
+	/* Info about slot being monitored for worker's naptime purpose */
+	struct SlotSyncWorkerWatchSlot
+	{
+		NameData	slot_name;
+		XLogRecPtr	confirmed_lsn;
+		int			inactivity_count;
+	}			monitoring_info;
+
+} SlotSyncWorker;
+
 /*
  * State of the transaction in parallel apply worker.
  *
@@ -234,12 +279,14 @@ extern PGDLLIMPORT struct WalReceiverConn *LogRepWorkerWalRcvConn;
 /* Worker and subscription objects. */
 extern PGDLLIMPORT Subscription *MySubscription;
 extern PGDLLIMPORT LogicalRepWorker *MyLogicalRepWorker;
+extern PGDLLIMPORT SlotSyncWorker *MySlotSyncWorker;
 
 extern PGDLLIMPORT bool in_remote_transaction;
 
 extern PGDLLIMPORT bool InitializingApplyWorker;
 
 extern void logicalrep_worker_attach(int slot);
+extern void slotsync_worker_attach(int slot);
 extern LogicalRepWorker *logicalrep_worker_find(Oid subid, Oid relid,
 												bool only_running);
 extern List *logicalrep_workers_find(Oid subid, bool only_running);
diff --git a/src/include/storage/lwlock.h b/src/include/storage/lwlock.h
index d77410bdea..334315acba 100644
--- a/src/include/storage/lwlock.h
+++ b/src/include/storage/lwlock.h
@@ -207,6 +207,7 @@ typedef enum BuiltinTrancheIds
 	LWTRANCHE_PGSTATS_DATA,
 	LWTRANCHE_LAUNCHER_DSA,
 	LWTRANCHE_LAUNCHER_HASH,
+	LWTRANCHE_SLOTSYNC_DSA,
 	LWTRANCHE_FIRST_USER_DEFINED
 }			BuiltinTrancheIds;
 
diff --git a/src/tools/pgindent/typedefs.list b/src/tools/pgindent/typedefs.list
index f2af84d7ca..9dff93f70e 100644
--- a/src/tools/pgindent/typedefs.list
+++ b/src/tools/pgindent/typedefs.list
@@ -1428,6 +1428,7 @@ LimitState
 LimitStateCond
 List
 ListCell
+ListDBForLogicalSlotsCmd
 ListDictionary
 ListParsedLex
 ListenAction
@@ -2304,6 +2305,7 @@ RelocationBufferInfo
 RelptrFreePageBtree
 RelptrFreePageManager
 RelptrFreePageSpanLeader
+RemoteSlot
 RenameStmt
 ReopenPtrType
 ReorderBuffer
@@ -2559,6 +2561,7 @@ SlabBlock
 SlabContext
 SlabSlot
 SlotNumber
+SlotSyncWorker
 SlruCtl
 SlruCtlData
 SlruErrorCause
@@ -3005,6 +3008,7 @@ WalLevel
 WalRcvData
 WalRcvExecResult
 WalRcvExecStatus
+WalRcvRepSlotDbData
 WalRcvState
 WalRcvStreamOptions
 WalRcvWakeupReason
-- 
2.34.1

